# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Solve the initial value problem of a first-order ordinary differential equation (ODE) system

    w'     = f(w,t)
    w(t=0) = w0      (initial condition)

where f is an arbitrary user-supplied function.

See PythonKernel and CythonKernel in pydgq.solver.kernel_interface, and pydgq.solver.builtin_kernels
for examples of implementing Cython kernels).

The main point of interest in this module is the function ivp(). Some auxiliary routines are also
provided.
"""

from __future__ import division, print_function, absolute_import

#from libc.stdlib cimport malloc, free

# use fast math functions from <math.h>, available via Cython
#from libc.math cimport fabs as c_abs
#from libc.math cimport sin, cos, log, exp, sqrt

#import cython

import numpy as np

cimport pydgq.solver.fputils as fputils    # nan/inf checks for arrays
from pydgq.solver.cminmax cimport cuimax  # "C unsigned int max"

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.types import DTYPE, RTYPE, DNAN

from pydgq.solver.kernel_interface cimport KernelBase  # RHS f() for w' = f(w,t)

from pydgq.solver.integrator_interface cimport IntegratorBase  # interface for integrator algorithms
cimport pydgq.solver.explicit as explicit  # classical explicit integrators
import  pydgq.solver.explicit as explicit
cimport pydgq.solver.implicit as implicit  # classical implicit integrators
import  pydgq.solver.implicit as implicit
cimport pydgq.solver.galerkin as galerkin  # Galerkin integrators
import  pydgq.solver.galerkin as galerkin

### Tell Cython that GCC's __float128 behaves like a double
### (this only concerns the Cython to C compile process and doesn't generate an actual C typedef)
###
### (There is also np.float128, which we could use.)
###
##cdef extern from "math.h":
##    ctypedef double __float128

#cdef extern from "complex.h":
#    double creal(double complex z) nogil
#    double cimag(double complex z) nogil
#    double complex conj(double complex z) nogil


#########################################################################################
# End-of-timestep boilerplate
#########################################################################################

# Store final value from this timestep to result array, if we have passed the start-of-recording point.
#
# This covers the most common case where interp=1 (save value at end of each timestep only).
#
# Galerkin integrators have their own implementation that accounts for interp.
#
# - t is the time value at the point where the solution is being stored.
# - wrk must have space for n_space_dofs items.
#
cdef inline void store( DTYPE_t* w, int n_space_dofs, int timestep, RTYPE_t t, int save_from, DTYPE_t* ww, RTYPE_t* tt, KernelBase rhs, DTYPE_t* ff, int* pfail, int failure, DTYPE_t* wrk ) nogil:
    cdef unsigned int n, j
    cdef DTYPE_t* wp = wrk

    # Note indexing:
    #  - save_from=1 means "save from first timestep onward" (1-based index 1, although we use 0-based indices).
    #  - save_from=0 means that also the initial condition is saved into the results.
    #
    if timestep >= save_from:
        # Compute the output slot number.
        n = timestep - save_from

        # Write output.
        tt[n] = t
        for j in range(n_space_dofs):
            ww[n*n_space_dofs + j] = w[j]

        # Optionally output the time derivative of the state vector (obtained via f()).
        #
        # This takes some extra time due to the extra rhs.call(), but can be useful for some visualizations.
        #
        if ff:
            rhs.begin_iteration(-1)  # iteration -1 = evaluating final result from this timestep
            rhs.call(w, wp, t)
            for j in range(n_space_dofs):
                ff[n*n_space_dofs + j] = wp[j]

        # Save failure flag if an array was provided for this.
        #
        if pfail:
            pfail[n] = failure


#########################################################################################
# Helper functions for modules using odesolve
#########################################################################################

def n_saved_timesteps( nt, save_from ):
    """def n_saved_timesteps( nt, save_from ):

Determine number of timesteps that will be saved.

Note that this is **not** the length of the result array; for that, see result_len().
The number returned by this function matches the length of the output arrays from timestep_boundaries().

Parameters:
    See result_len().

Returns:
    int:
        the number of timesteps that will be saved.
"""
    if nt < 1:
        raise ValueError( "nt must be >= 1, got %d" % (nt) )
    if save_from < 0:
        raise ValueError( "save_from must be >= 0, got %d" % (save_from) )
    if save_from > nt:
        raise ValueError( "save_from must be <= nt, otherwise nothing to do; got save_from = %d, nt = %d" % (save_from, nt) )

    n = nt - (max(1, save_from) - 1)  # this many actual timesteps will be saved
    if save_from == 0:
        return n + 1  # the initial condition takes one output slot
    else:
        return n


def result_len( int nt, int save_from, int interp=1 ):
    """def result_len( int nt, int save_from, int interp=1 ):

Determine length of storage needed on the time axis for ivp().

Parameters:
    nt : int, >= 1
        Number of timesteps to take.

    save_from : int, >= 0
        Index of first timestep to save.

        This allows discarding part of the data at the beginning.

        The special value 0 means that also the initial condition
        will be copied into the results. The initial condition
        always produces exactly one item.

        A value of 1 means "save results from first actual timestep onward";
        likewise for higher values (e.g. 2 -> second timestep onward).

    interp : int, >= 1
        Galerkin integrators, such as dG, have the possibility of
        evaluating the solution at points inside the timestep,
        by evaluating the computed Galerkin approximation.

        This sets the number of result points that will be generated
        per computed timestep.

        The maximum allowed value for interp is galerkin.datamanager.maxnt_vis
        (initialized when galerkin.init() loads its data file).

        interp=1 means that only the value at the endpoint
        of each timestep will be saved.

        For all non-Galerkin integrators, interp=1 is the only valid setting.
        This is because the other integrators are based on collocation methods
        (point-based methods), so the value of the solution at points other
        than the timestep boundaries is undefined for them.

        For dG:
            A value of interp >= 2 means that n values equally spaced in time,
            from the start of the timestep to its end, will be saved.

            Examples:
              - interp=2 saves the start and end values for each timestep.
              - interp=3 saves these and also the midpoint value.
              - interp=11 gives a nice spacing of 10 equal intervals
                (i.e. 11 values including the fencepost) across each timestep.

            Note that in **discontinuous** Galerkin,
            the start value of a timestep will be different
            from the end value of the previous timestep,
            although these share the same time coordinate!

            Mathematically, the solution is defined to be left-continuous,
            i.e., the "end value" is the actual value. The "start value"
            is actually a one-sided limit from the right, taken toward
            the start of the timestep.

        For cG:
            In **continuous** Galerkin, the solution is continuous,
            so the endpoint of timestep n is the start point of timestep n+1.

            The values are equally spaced, but avoiding the duplicate.
            Effectively this takes the visualization points for interp
            one larger than specified, and discards the first one.

            Examples:
              - interp=2 saves the midpoint and end values for each timestep.
              - interp=4 saves values at relative offsets dt/4,
                dt/2, 3dt/4 and dt from the timestep start.
                There are effectively five "fenceposts" and
                four intervals in each timestep [n*dt, (n+1)*dt].

Returns:
    int:
        Number of storage slots (i.e. array length) along the time axis,
        that are needed by ivp(), when called with the given parameter values.
"""
    if nt < 1:
        raise ValueError( "nt must be >= 1, got %d" % (nt) )
    if save_from < 0:
        raise ValueError( "save_from must be >= 0, got %d" % (save_from) )
    if save_from > nt:
        raise ValueError( "save_from must be <= nt, otherwise nothing to do; got save_from = %d, nt = %d" % (save_from, nt) )
    if interp < 1:
        raise ValueError( "interp must be >= 1, got %d" % (interp) )

    if save_from == 0:
        # 1 = the initial condition
        return int( 1 + nt*interp )
    else:
        # save_from = 1 means that the initial condition is not saved,
        # but the first timestep (n=0 in the timestep loop) and onward are saved.
        return int( (nt - (save_from - 1))*interp )


def timestep_boundaries( int nt, int save_from, int interp=1 ):
    """def timestep_boundaries( int nt, int save_from, int interp=1 ):

Return start and one-past-end indices for each timestep in the result.

These can be used to index the arrays tt, ww and ff (see ivp()) on the time axis.

This is useful with Galerkin integrators, which support several visualization points per timestep (interp > 1).

Parameters:
    See result_len().

Returns:
    Tuple (startj, endj):
        startj (respectively endj) is a rank-1 np.array containing the start (resp. one-past-end) indices for each timestep.

        The indices for timestep n are range(startj[n], endj[n]).

        If save_from == 0, the initial condition (always exactly one point) counts as "timestep 0"; otherwise "timestep 0" is the first saved timestep.
"""
    if nt < 1:
        raise ValueError( "nt must be >= 1, got %d" % (nt) )
    if save_from < 0:
        raise ValueError( "save_from must be >= 0, got %d" % (save_from) )
    if save_from > nt:
        raise ValueError( "save_from must be <= nt, otherwise nothing to do; got save_from = %d, nt = %d" % (save_from, nt) )
    if interp < 1:
        raise ValueError( "interp must be >= 1, got %d" % (interp) )

    cdef unsigned int n, offs
    cdef unsigned int n_saved_timesteps = nt - (cuimax(1, save_from) - 1)
    cdef unsigned int n_output          = n_saved_timesteps

    if save_from == 0:
        n_output += 1  # the initial condition takes one output slot

    cdef int[::1] startj = np.empty( [n_output], dtype=np.intc, order="C" )
    cdef int[::1] endj   = np.empty( [n_output], dtype=np.intc, order="C" )  # one-past-end

    if save_from == 0:
        offs = 1  # one output slot was taken; shift the rest

        # The initial condition always produces exactly one point
        startj[0] = 0
        endj[0]   = 1
    else:
        offs = 0

    with nogil:
        for n in range(n_saved_timesteps):
            # Loop over visualization points in the timestep.
            #
            startj[offs+n] = offs           + n*interp
            endj[offs+n]   = startj[offs+n] +   interp  # actually one-past-end

    return (np.asanyarray(startj), np.asanyarray(endj))


#########################################################################################
# Integrator
#########################################################################################

# TODO: add convergence tolerance (needs some changes in implicit.pyx and galerkin.pyx (basically wherever "maxit" is used))
def ivp( str integrator, int allow_denormals, DTYPE_t[::1] w0 not None, RTYPE_t dt, int nt, int save_from, int interp,
         KernelBase rhs, DTYPE_t[:,::1] ww, DTYPE_t[:,::1] ff, int[::1] fail, RTYPE_t RK2_beta=1.0,
         int maxit=100 ):
    """def ivp( str integrator, int allow_denormals, DTYPE_t[::1] w0 not None, RTYPE_t dt, int nt, int save_from, int interp,
         KernelBase rhs, DTYPE_t[:,::1] ww, DTYPE_t[:,::1] ff, int[::1] fail, RTYPE_t RK2_beta=1.0,
         int maxit=100 ):

Solve initial value problem.

This routine integrates first-order ordinary differential equation (ODE) systems of the form

    w'     = f(w, t)
    w(t=0) = w0       (initial condition)

where f is a user-provided kernel for computing the RHS.

Parameters:

    integrator : str
        Time integration algorithm. One of:

            SE : Symplectic Euler (also known as semi-implicit Euler)
                1st order accuracy, symplectic, conserves energy approximately,
                very fast, may require a smaller timestep than the others.

                !!! Only for second-order problems which have been reduced
                    to first-order form. See the user manual for details. !!!

            BE : Backward Euler (implicit Euler).
                1st order accuracy, high numerical dissipation.

                A-stable for linear problems, but due to implementation constraints,
                in this nonlinear solver arbitrarily large timesteps cannot be used.

                (The timestep size is limited by the loss of contractivity in the
                 Banach iteration as the timestep becomes larger than some
                 situation-specific critical value.)

            IMR : Implicit Midpoint Rule.
                2nd order accuracy, symplectic, conserves energy approximately.
                Slow, but may work with a larger timestep than others.

            RK4 : 4th order Runge-Kutta.
                4th order accuracy, but not symplectic and does not conserve energy;
                computed orbits may drift arbitrarily far from the true orbits
                in a long simulation (esp. in a vibration simulation). Moderately fast.

            RK3 : Kutta's third-order method.

            RK2 : parametric second-order Runge-Kutta.
                Takes the optional parameter RK2_beta, which controls where inside the timestep
                the second evaluation of f() is taken.

                RK2_beta must be in the half-open interval (0, 1]. Very small values
                will cause problems (beta appears in the denominator in the final summation formula).

                Popular choices:
                    beta = 1/2          , explicit midpoint method
                    beta = 2/3          , Ralston's method
                    beta = 1   (default), Heun's method, also known as the explicit trapezoid rule

            FE : Forward Euler (explicit Euler).
                1st order accuracy, very unstable, requires a very small timestep.

                Provided for reference only.

            dG : discontinuous Galerkin (recommended)
                An advanced implicit method. Finds a weak solution that is finitely
                discontinuous (C^{-1}) across timestep boundaries.
                Typically works with large-ish timesteps.

                The solution satisfies the Galerkin orthogonality property:
                the residual of the result is L2-orthogonal to the basis functions.
                Roughly, this means that the numerical solution is, in the least-squares sense,
                the best representation (in the given basis) of the unknown true solution.

                See galerkin.init() for configuration (polynomial degree of basis).

            cG : continuous Galerkin.
                Like dG, but the solution is C^0 continuous across timestep boundaries.

                See galerkin.init() for configuration (polynomial degree of basis).

    allow_denormals : bool
        If True,  allow denormal numbers in computation (10...100x performance penalty
                  on most modern processors).

        If False, stop computation and fill the rest of the solution with zeroes
                  when denormal numbers are reached.

        Denormal numbers are usually encountered in cases with high damping (beta) and
        low external load (mu_m, Pi_F); in such a case the system quickly spirals in onto
        the equilibrium point at the origin, with the magnitude of the numbers decreasing
        very rapidly.

        Setting this to True gives a more beautiful plot in such cases, but that can be
        traded for a large performance increase by setting this to False.

        In cases where denormal numbers are not encountered, this option has no effect
        on the results.

    w0 : rank-1 np.array
        initial state (w1, w2, ..., wn).
        This also automatically sets n_space_dofs (for the current problem) to len(w0).

    dt : RTYPE_t, != 0
        timestep size

        Negative values can be used to integrate backward in time.

        !!! The time t always starts at zero; if your RHS explicitly depends on t,
            take this into account! !!!

    nt : int, >= 1
        number of timesteps to take

    save_from : int, >= 0, <= nt
        first timestep index to save, 0-based (allows discarding part of the data at the beginning)

        0 = save initial condition and all timesteps in results
        1 = save all timesteps (but don't save initial condition)
        2 = save from second timestep onward
        ...

    interp : int
        For Galerkin methods: how many visualization points to produce per computed timestep.
        For all other integrators, interp must be 1.

    rhs : instance of class derived from KernelBase
        Kernel implementing the right-hand side of  u' = f(u, t)

    ww : DTYPE_t[:,::1] of size [result_len(),n_space_dofs] or None
        Output array for w.
         If None:
            Will be created and returned.
         If supplied:
            The user-given array will be filled in. No bounds checking - make sure it is large enough!

    ff : DTYPE_t[:,::1] of size [result_len(),n_space_dofs] or None
        If not None, output array for w' (the time derivative of w).

    fail : int[::1] of size [n_saved_timesteps(),] or None
        (NOTE: size on axis 0 different from that of ww and ff! One entry per timestep, regardless of interp.)

        If not None, output array for status flag for each timestep:
            0 = converged to machine precision
            1 = did not converge to machine precision

        This data is only meaningful for implicit methods (IMR, BE, dG, cG); explicit methods will simply flag success for each timestep.

        If save_from == 0, the initial condition counts as the zeroth timestep, and is always considered as converged.

    maxit : int, >= 1
        Maximum number of Banach/Picard iterations to take at each timestep.

        Only meaningful if an implicit integrator is used (BE, IMR, dG, cG).

Return value:
    tuple (ww, tt):
        ww : DTYPE_t[:,::1] of size [result_len(),n_space_dofs]
            Solution.
            If input ww is not None, it is passed through here.
            If input ww is None, the created array is returned here.
        tt : DTYPE_t[::1] of size [result_len()]
            Time values corresponding to the solution values in ww.
"""
    # Parameter validation
    #
    known_integrators    = ["IMR", "BE", "RK4", "RK3", "RK2", "FE", "dG", "cG", "SE"]
    galerkin_integrators = ["dG", "cG"]  # integrators, of those already listed in known_integrators, which are based on Galerkin methods.

    if integrator not in known_integrators:
        raise ValueError("Unknown integrator '%s'; valid: %s" % ( integrator, ", ".join(known_integrators) ))

    if integrator not in galerkin_integrators and interp != 1:
        raise ValueError("For non-Galerkin integrators (such as the chosen integrator='%s'), interp must be 1, but interp=%d was given." % (integrator, interp))

    if dt == 0.0:
        raise ValueError( "dt cannot be zero" )
    if nt < 1:
        raise ValueError( "nt must be >= 1, got %d" % (nt) )
    if save_from < 0:
        raise ValueError( "save_from must be >= 0, got %d" % (save_from) )
    if save_from > nt:
        raise ValueError( "save_from must be <= nt, otherwise nothing to do; got save_from = %d, nt = %d" % (save_from, nt) )
    if interp < 1:
        raise ValueError( "interp must be >= 1, got %d" % (interp) )
    if maxit < 1:
        raise ValueError( "maxit must be >= 1, got %d" % (maxit) )

    cdef int n_slots      = result_len( nt, save_from, interp )
    cdef int n_space_dofs = w0.shape[0]
    if ww is not None:
        if ww.shape[0] != n_slots:
            raise ValueError( "shape of output array ww not compatible with length of output: shape(ww)[0] = %d, but %d values are to be saved" % (ww.shape[0], n_slots) )
        if ww.shape[1] != n_space_dofs:
            raise ValueError( "shape of output array ww not compatible with n_space_dofs: shape(ww)[1] = %d, but n_space_dofs = %d" % (ww.shape[1], n_space_dofs) )

    if ff is not None:
        if ff.shape[0] != n_slots:
            raise ValueError( "shape of output array ff not compatible with length of output: shape(ff)[0] = %d, but %d timesteps are to be saved" % (ff.shape[0], n_slots) )
        if ff.shape[1] != n_space_dofs:
            raise ValueError( "shape of output array ff not compatible with n_space_dofs: shape(ff)[1] = %d, but n_space_dofs = %d" % (ff.shape[1], n_space_dofs) )

    # Runtime sanity checking of the result
    #
    cdef int do_denormal_check  = <int>(not allow_denormals)  # if denormals not allowed, check them
    cdef int denormal_triggered = 0
    cdef int naninf_triggered   = 0

    # output: w
    #
    if ww is None:
        ww = np.empty( (n_slots,n_space_dofs), dtype=DTYPE, order="C" )
    cdef DTYPE_t* pww = &ww[0,0]

    # output: time values
    #
    cdef RTYPE_t[::1] tt = np.empty( (n_slots,), dtype=RTYPE, order="C" )
    cdef RTYPE_t* ptt = &tt[0]

    # optional output: w'
    #
    # ("ff" because it is the value of the RHS f())
    #
    cdef DTYPE_t* pff
    if ff is not None:
        pff = &ff[0,0]
    else:
        pff = <DTYPE_t*>0  # store() knows to omit saving w' if the pointer is NULL

    # optional output: status flag, one per timestep
    #
    cdef int* pfail
    if fail is not None:
        pfail = &fail[0]
    else:
        pfail = <int*>0

    # Initial condition: initialize the value of w at the end of the previous timestep to the initial condition of the problem.
    #
    cdef DTYPE_t[::1] w_arr = np.empty( (n_space_dofs,), dtype=DTYPE, order="C" )
    cdef DTYPE_t* w = &w_arr[0]  # we only need a raw pointer
    cdef unsigned int j
    for j in range(n_space_dofs):
        w[j] = w0[j]

    # Fill in stuff from the initial condition to the results, if saving all the way from the start.
    #
    # Temporary storage for w' as output by rhs.call(). This is needed later anyway, but we may need it already here.
    cdef DTYPE_t[::1] wp_arr = np.empty( (n_space_dofs,), dtype=DTYPE, order="C" )
    cdef DTYPE_t* wp = &wp_arr[0]
    if save_from == 0:
        # State vector w
        ww[0,:] = w0
        tt[0]   = 0.

        # w'
        if ff is not None:
            rhs.begin_timestep(0)  # timestep 0 = initial condition
            rhs.begin_iteration(-1)  # iteration -1 = evaluating final result from this timestep
            rhs.call(&w0[0], wp, 0.0)  # t at beginning = 0.0
            for j in range(n_space_dofs):
                ff[0,j] = wp[j]

        # success/fail information (initial condition is always successful)
        if fail is not None:
            fail[0] = 0

    # Timestep number and current time
    #
    cdef unsigned int n = 0
    cdef RTYPE_t t = 0.0

    # Work space for store()
    #
    cdef DTYPE_t[::1] wrk_arr = np.empty( (n_space_dofs,), dtype=DTYPE, order="C" )
    cdef DTYPE_t* wrk = &(wrk_arr[0])

    # Implicit methods support
    #
    cdef unsigned int nits                     # number of implicit iterations (Banach fixed point iterations) taken at this timestep
    cdef unsigned int nfail         = 0        # number of last failed timestep (failed = did not converge to machine precision)
    cdef unsigned int totalfailed   = 0        # how many failed timesteps in total
    cdef unsigned int totalnits     = 0        # total number of iterations taken, across all timesteps
    cdef unsigned int max_taken_its = 0        # maximum number of iterations (seen) that was taken for one timestep
    cdef unsigned int min_taken_its = 2*maxit  # minimum number of iterations (seen) that was taken for one timestep. (Invalid value used to force initialization at first timestep.)

    # Galerkin methods support
    #
    cdef unsigned int offs, out_start, out_end, l, noutput

    # Integrator object
    #
    cdef IntegratorBase algo
    cdef galerkin.GalerkinIntegrator algo_g

    # Account for the output slot possibly used for the initial condition.
    #
    # We need this for Galerkin methods, but also for filling the empty slots when the solver fails.
    #
    if save_from == 0:
        offs = 1
    else:
        offs = 0


    # Integration loop
    #
    if integrator in ["SE", "RK4", "RK3", "RK2", "FE", "BE", "IMR"]:  # classical integrators
        # explicit integrators
        if integrator == "SE":  # symplectic Euler
            if n_space_dofs % 2 != 0:
                raise ValueError("SE: Symplectic Euler (SE) only makes sense for second-order systems transformed to first-order ones, but got odd number of n_space_dofs = %d" % (n_space_dofs))
            algo = explicit.SE(rhs)
        elif integrator == "RK4":
            algo = explicit.RK4(rhs)  # classical fourth-order Runge-Kutta
        elif integrator == "RK3":
            algo = explicit.RK3(rhs)  # Kutta's third-order method
        elif integrator == "RK2":
            algo = explicit.RK2(rhs, RK2_beta)  # parametric second-order Runge-Kutta
        elif integrator == "FE":
            algo = explicit.FE(rhs)   # forward Euler
        # implicit integrators
        elif integrator == "BE":
            algo = implicit.BE(rhs, maxit)  # backward Euler
        else: # integrator == "IMR":
            algo = implicit.IMR(rhs, maxit)  # implicit midpoint rule

        # We release the GIL for the integration loop to let another Python thread execute
        # while this one is running through a lot of timesteps (possibly several million per solver run).
        #
        with nogil:
            for n in range(1,nt+1):
                t = (n-1)*dt  # avoid accumulating error (don't sum; for very large t, this version will tick as soon as the floating-point representation allows it)

                rhs.begin_timestep(n)
                nits = algo.call( w, t, dt )

                # update the iteration statistics
                #
                # (technically, only the implicit algorithms really need this)
                #
                if nits == maxit:
                    totalfailed += 1
#                if nfail == 0  and  nits == maxit:  # store first failed timestep
                if nits == maxit:  # store last failed timestep (update at each, last one left standing)
                    nfail = n
                if nits > max_taken_its:
                    max_taken_its = nits
                if nits < min_taken_its:
                    min_taken_its = nits
                totalnits += nits

                # end-of-timestep boilerplate
                #
                t = n*dt
                store( w, n_space_dofs, n, t, save_from, pww, ptt, rhs, pff, pfail, <int>(nits == maxit), wrk )
                if do_denormal_check:
                    # In practice this check seems good enough for an IVP solver, although it does run the theoretical risk
                    # of triggering (with a nonzero probability) when the solution just passes through zero.
                    #
                    # To be more sure, we could check the l-infinity (max abs) vector norm of the solution,
                    # and see if it has been decreasing for N timesteps (for some suitable N) before
                    # declaring the rest of the solution as denormal.
                    #
                    denormal_triggered = fputils.all_denormal( w, n_space_dofs )
                naninf_triggered = fputils.any_naninf( w, n_space_dofs )
                if denormal_triggered or naninf_triggered:
                    break

        nt_taken = max(1, n)
#        # DEBUG
#        failed_str = "" if totalfailed == 0 else "; last non-converged timestep %d" % (nfail)
#        print( "    min/avg/max iterations taken = %d, %g, %d; total number of non-converged timesteps %d (%g%%)%s" % (int(min_taken_its), float(totalnits)/nt_taken, int(max_taken_its), totalfailed, 100.0*float(totalfailed)/nt_taken, failed_str) )

    else:  # integrator in galerkin_integrators:  # Galerkin integrators
        # TODO: this is now almost the same code as above; only the custom store() logic for Galerkin integrators needs this different branch.

        # NOTE: instantiating DG or CG also checks whether galerkin.datamanager exists and is initialized
        #
        if integrator == "dG":  # discontinuous Galerkin (recommended!)
            algo = galerkin.DG(rhs, maxit)
        else: # integrator == "cG":  # continuous Galerkin (doesn't work that well in practice)
            algo = galerkin.CG(rhs, maxit)

        # typecast to force Cython to see private stuff defined in GalerkinIntegrator (TODO/FIXME: leaky interface, the Galerkin equivalent of store() belongs in galerkin.pyx)
        algo_g = <galerkin.GalerkinIntegrator>algo

        # final sanity check
        #
        if galerkin.datamanager.nt_vis != interp:  # TODO: relax this implementation-technical limitation
            raise NotImplementedError("%s: interp = %d, but galerkin.init() was last called with different nt_vis = %d; currently this is not supported." % (integrator, interp, galerkin.datamanager.nt_vis))

        with nogil:
            for n in range(1,nt+1):
                t = (n-1)*dt

                rhs.begin_timestep(n)
                nits = algo.call( w, t, dt )

                # update the iteration statistics
                #
                if nits == maxit:
                    totalfailed += 1
#                if nfail == 0  and  nits == maxit:  # store first failed timestep
                if nits == maxit:  # store last failed timestep (update at each, last one left standing)
                    nfail = n
                if nits > max_taken_its:
                    max_taken_its = nits
                if nits < min_taken_its:
                    min_taken_its = nits
                totalnits += nits

                # end-of-timestep boilerplate
                #
                if interp == 1:
                    # In this case, we already have what we need. Just do what the other methods do,
                    # saving the end-of-timestep value of the solution from w.
                    #
                    t = n*dt
                    store( w, n_space_dofs, n, t, save_from, pww, ptt, rhs, pff, pfail, <int>(nits == maxit), wrk )
                else:
                    # Inline custom store() implementation accounting for interp (and how to assemble a Galerkin series)
                    #
                    if n >= save_from:
                        # Interpolate inside timestep using the Galerkin representation of the solution, obtaining more visualization points.
                        #
                        algo_g.assemble( algo_g.psivis, algo_g.uvis, algo_g.ucvis, interp )  # basis, output, wrk, n_points
                        noutput = n - cuimax(1, save_from)  # 0-based timestep number starting from the first saved one.
                                                                    # Note that n = 1, 2, ... (also store() depends on this numbering!)
                        out_start = offs + noutput*interp
                        for l in range(interp):
                            t = ((n-1) + algo_g.tvis[l]) * dt
                            ptt[(out_start+l)] = t
                            for j in range(n_space_dofs):
                                # uvis: [nt_vis,n_space_dofs]
                                pww[(out_start+l)*n_space_dofs + j] = algo_g.uvis[l*n_space_dofs + j]

                        # Optionally output the time derivative of the state vector (obtained via f()).
                        #
                        # This takes some extra time, but can be useful for some visualizations.
                        #
                        if pff:
                            rhs.begin_iteration(-1)  # iteration -1 = evaluating final result from this timestep
                            for l in range(interp):
                                t = ((n-1) + algo_g.tvis[l]) * dt
                                rhs.call(&algo_g.uvis[l*n_space_dofs + 0], wp, t)
                                for j in range(n_space_dofs):
                                    pff[(out_start+l)*n_space_dofs + j] = wp[j]

                        if pfail:
                            pfail[offs + noutput] = <int>(nits == maxit)

                if do_denormal_check:
                    denormal_triggered = fputils.all_denormal( w, n_space_dofs )
                naninf_triggered = fputils.any_naninf( w, n_space_dofs )
                if denormal_triggered or naninf_triggered:
                    break

        nt_taken = max(1, n)

#        # DEBUG
#        failed_str = "" if totalfailed == 0 else "; last non-converged timestep %d" % (nfail)
#        print( "    min/avg/max iterations taken = %d, %g, %d; total number of non-converged timesteps %d (%g%%)%s" % (int(min_taken_its), float(totalnits)/nt_taken, int(max_taken_its), totalfailed, 100.0*float(totalfailed)/nt_taken, failed_str) )


#    # DEBUG/INFO: final value of w'
#    #
#    t = nt_taken*dt  # time at start of "next" timestep
#    rhs.begin_timestep(nt_taken+1)
#    rhs.begin_iteration(-1)  # iteration -1 = evaluating final result from this timestep
#    rhs.call(w, wp, t)
#    lw  = [ "%0.18g" % w[j] for j in range(n_space_dofs) ]
#    sw  = ", ".join(lw)
#    lwp = [ "%0.18g" % wp[j] for j in range(n_space_dofs) ]
#    swp = ", ".join(lwp)
#    print( "    final w = %s\n    final f(w) = %s" % (sw, swp) )

    # If a failure check triggered, mark the rest of the solution accordingly.
    #
    cdef DTYPE_t fill
    cdef int failflag
    if denormal_triggered  or  naninf_triggered:
        # how many timesteps were saved to the output array
        if n < save_from:
            noutput = 0
        else:
            noutput = n - cuimax(1, save_from)  # 0-based timestep number starting from the first saved one.
                                                # Note that n = 1, 2, ... (also store() depends on this numbering!)

        # first unfilled slot in output array
        out_start = offs + noutput*interp

        if denormal_triggered:
            fill = 0.0  # w small, no more change in w  =>  both w = 0 and w' = 0
            failflag = 0  # successful
        else: # naninf_triggered:
            fill = DNAN
            failflag = 1

        ww[out_start:,:] = fill
        tt[out_start:]   = np.nan  # always real-valued NaN
        if ff is not None:
            ff[out_start:,:] = fill
        if fail is not None:
            fail[(offs + noutput):] = failflag  # no interp in flag array

    return (np.asanyarray(ww), np.asanyarray(tt))




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.kernel_interface cimport KernelBase
from pydgq.solver.integrator_interface cimport ExplicitIntegrator

# Fourth-order Runge-Kutta
cdef class RK4(ExplicitIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# Kutta's third-order method
cdef class RK3(ExplicitIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# Parametric second-order Runge-Kutta
cdef class RK2(ExplicitIntegrator):
    cdef RTYPE_t beta  # saved from __init__ (def method, not declared here)
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# Forward Euler (not recommended!)
cdef class FE(ExplicitIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# Symplectic Euler (for 2nd-order problems reduced to a twice larger 1st-order system)
cdef class SE(ExplicitIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil




# -*- coding: utf-8 -*-
#
# Tests/usage examples for kernels built-in to the solver.

from __future__ import division, print_function

import numpy as np

import matplotlib.pyplot as plt

from pydgq.solver.types import DTYPE
from pydgq.solver.galerkin import init
import pydgq.solver.builtin_kernels
import pydgq.solver.odesolve
from pydgq.utils.discontify import discontify  # for plotting dG results


#####################
# config for testing
#####################

q = 2  # degree of basis for dG and cG

# How many visualization (interpolation) points to use within each timestep for Galerkin methods.
#
# Note that the dG solution has the best accuracy at the endpoint of the timestep;
# to compare apples-to-apples with classical integrators, this should be set to 1.
#
# Larger values (e.g. 11) are useful for visualizing the behavior of the dG solution inside
# the timestep (something the classical integrators do not model at all).
#
nt_vis_galerkin = 1

nt = 100   # number of timesteps
dt = 0.02  # timestep size

save_from = 0  # see pydgq.solver.odesolve.ivp()


#####################
# main program
#####################

# rel_tol : how close the numerical solution must be to the exact analytical one, in l-infinity norm (max abs)
def test(integrator, nt_vis, rel_tol=1e-2, vis=False):
    n_saved_timesteps = pydgq.solver.odesolve.n_saved_timesteps( nt, save_from )
    result_len        = pydgq.solver.odesolve.result_len( nt, save_from, interp=nt_vis )
    startj,endj       = pydgq.solver.odesolve.timestep_boundaries( nt, save_from, interp=nt_vis )

    n   = 4  # number of DOFs in the **1st-order** system

    # set IC
    w0  = np.empty( (n,), dtype=DTYPE, order="C" )
    w0[0] =  0.  # u1
    w0[1] =  1.  # v1
    w0[2] =  2.  # u2
    w0[3] = -1.  # v2

    # set up the M0, M1 and M2 matrices for  M2 u'' = M0 u + M1 u'
    #
    m  = n//2
    M0 = -np.eye(   m,     dtype=DTYPE)
    M1 =  np.zeros( (m,m), dtype=DTYPE)

    M2 = np.zeros( (m,m), dtype=DTYPE, order="C" )
    M2[0,1] = 1.
    M2[1,0] = 1.

    # for checking the result
    invM2_times_M0 = (np.linalg.inv(M2)).dot(M0)
    invM2_times_M1 = (np.linalg.inv(M2)).dot(M1)

    # instantiate kernel
    rhs1 = pydgq.solver.builtin_kernels.Linear2ndOrderKernelWithMassMatrix(n, M0, M1, M2)        # note n, not m
    rhs2 = pydgq.solver.builtin_kernels.Linear2ndOrderKernel(n, invM2_times_M0, invM2_times_M1)  # different algorithm, but should give the same result

    # create output arrays
    ww   = None #np.empty( (result_len,n), dtype=DTYPE, order="C" )    # result array for w; if None, will be created by ivp()
    ff   = np.empty( (result_len,n), dtype=DTYPE, order="C" )          # optional,  result array for w', could be None
    fail = np.empty( (n_saved_timesteps,), dtype=np.intc, order="C" )  # optional,  fail flag for each timestep, could be None

    # solve problem
    ww,tt = pydgq.solver.odesolve.ivp( integrator=integrator, allow_denormals=False,
                                       w0=w0, dt=dt, nt=nt,
                                       save_from=save_from, interp=nt_vis,
                                       rhs=rhs1,
                                       ww=ww, ff=ff, fail=fail,
                                       maxit=100 )

    # check result
    ww_ref,dummy = pydgq.solver.odesolve.ivp( integrator=integrator, allow_denormals=False,
                                       w0=w0, dt=dt, nt=nt,
                                       save_from=save_from, interp=nt_vis,
                                       rhs=rhs2,
                                       ww=ww, ff=ff, fail=fail,
                                       maxit=100 )

    relerr_linfty = np.linalg.norm(ww - ww_ref, ord=np.inf) / np.linalg.norm(ww_ref, ord=np.inf)
    if (relerr_linfty < rel_tol).all():
        passed = 1
        if not vis:
            print("PASS, %03s, tol=% 7g, relerr=%g" % (integrator, rel_tol, relerr_linfty))
    else:
        passed = 0
        if not vis:
            print("FAIL, %03s, tol=% 7g, relerr=%g" % (integrator, rel_tol, relerr_linfty))

    # visualize if requested
    if vis:
        plt.figure(1)
        plt.clf()

        # show the discontinuities at timestep boundaries if using dG (and actually have something to draw within each timestep)
        if integrator == "dG" and nt_vis > 1:
            tt = discontify( tt, endj - 1, fill="nan" )
            for j in range(n):
                # we need the copy() to get memory-contiguous data for discontify() to process
                wtmp = discontify( ww_ref[:,j].copy(), endj - 1, fill="nan" )
                plt.plot( tt, wtmp, 'r--' )  # known exact solution

                wtmp = discontify( ww[:,j].copy(), endj - 1, fill="nan" )
                plt.plot( tt, wtmp, 'k-'  )  # numerical solution
        else:
            plt.plot( tt, ww_ref, 'r--' )  # known exact solution
            plt.plot( tt, ww,     'k-'  )  # numerical solution

        plt.grid(b=True, which="both")
        plt.axis("tight")
        plt.xlabel(r"$t$")
        plt.ylabel(r"$w(t)$")

    return passed


if __name__ == '__main__':
    print("** Testing integrators with 2nd-order linear kernel with mass matrix **")

    stuff_to_test = ( ("SE",  1e-1,  False),  # 2nd-order problem, also "SE" is applicable
                      ("IMR", 1e-3,  False),
                      ("BE",  1e-1,  False),
                      ("RK4", 1e-8,  False),
                      ("RK3", 1e-6,  False),
                      ("RK2", 1e-3,  False),
                      ("FE",  1e-1,  False),
                      ("dG",  1e-12, True ),
                      ("cG",  1e-4,  True )
                    )
    n_passed = 0

    for integrator,rel_tol,is_galerkin in stuff_to_test:
        if is_galerkin:
            nt_vis = nt_vis_galerkin  # visualization points per timestep in Galerkin methods
            init(q=q, method=integrator, nt_vis=nt_vis, rule=None)
        else:
            nt_vis = 1   # other methods compute only the end value for each timestep

        n_passed += test(integrator, nt_vis, rel_tol)
    print("** %d/%d tests passed **" % (n_passed, len(stuff_to_test)))

#    # DEBUG: draw something
#    #
#    nt_vis = nt_vis_galerkin
#    init(q=q, method="dG", nt_vis=nt_vis, rule=None)
#    test(integrator="dG", nt_vis=nt_vis, vis=True)
#    plt.show()




# -*- coding: utf-8 -*-
#
# fast inline min/max functions for C code in Cython.

from __future__ import absolute_import

from pydgq.solver.types cimport RTYPE_t

# real
#
cdef inline RTYPE_t cfmin(RTYPE_t a, RTYPE_t b) nogil:
    return a if a < b else b
cdef inline RTYPE_t cfmax(RTYPE_t a, RTYPE_t b) nogil:
    return a if a > b else b

# int
#
cdef inline int cimin(int a, int b) nogil:
    return a if a < b else b
cdef inline int cimax(int a, int b) nogil:
    return a if a > b else b

# unsigned int
#
cdef inline unsigned int cuimin(unsigned int a, unsigned int b) nogil:
    return a if a < b else b
cdef inline unsigned int cuimax(unsigned int a, unsigned int b) nogil:
    return a if a > b else b




# -*- coding: utf-8 -*-
#
# Floating-point flags detection. Available at Cython level only.
#
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True

from __future__ import division, print_function, absolute_import

from pydgq.solver.types cimport DTYPE_t

cdef extern from "math.h":
    # fpclassify is actually a macro, so it does not have a specific input type.
    # Here we tell Cython we would like to use it for whatever type DTYPE_t is a typedef for.
    int fpclassify(DTYPE_t x) nogil
    int FP_INFINITE
    int FP_NAN
    int FP_NORMAL
    int FP_SUBNORMAL
    int FP_ZERO

# Check whether all components of w are denormal
#
# n : number of components in w
#
cdef inline int all_denormal( DTYPE_t* w, int n ) nogil:
    cdef unsigned int n_denormal = 0
    cdef unsigned int j
    for j in range(n):
        if fpclassify(w[j]) == FP_SUBNORMAL:
            n_denormal += 1

    if n_denormal == n:
        return 1
    else:
        return 0

# Check whether any component of w is nan or inf
#
# n : number of components in w
#
cdef inline int any_naninf( DTYPE_t* w, int n ) nogil:
    cdef unsigned int j
    cdef int c
    for j in range(n):
        c = fpclassify(w[j])
        if c == FP_NAN  or  c == FP_INFINITE:
            return 1
    return 0




# -*- coding: utf-8 -*-
#
"""Setup script for compiling the Cython example.

Usage:
    python -m setup build_ext --inplace
"""

from __future__ import division, print_function, absolute_import

#########################################################
# Config
#########################################################

# choose build type here
#
build_type="optimized"
#build_type="debug"


#########################################################
# Init
#########################################################

# check for Python 2.7 or later
# http://stackoverflow.com/questions/19534896/enforcing-python-version-in-setup-py
import sys
if sys.version_info < (2,7):
    sys.exit('Sorry, Python < 2.7 is not supported')

import os

from setuptools import setup
from setuptools.extension import Extension

try:
    from Cython.Build import cythonize
except ImportError:
    sys.exit("Cython not found. Cython is needed to build the extension modules for pydgq.")


#########################################################
# Definitions
#########################################################

extra_compile_args_math_optimized    = ['-fopenmp', '-march=native', '-O2', '-msse', '-msse2', '-mfma', '-mfpmath=sse']
extra_compile_args_math_debug        = ['-fopenmp', '-march=native', '-O0', '-g']

extra_compile_args_nonmath_optimized = ['-O2']
extra_compile_args_nonmath_debug     = ['-O0', '-g']

extra_link_args_optimized    = ['-fopenmp']
extra_link_args_debug        = ['-fopenmp']


if build_type == 'optimized':
    my_extra_compile_args_math    = extra_compile_args_math_optimized
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_optimized
    my_extra_link_args            = extra_link_args_optimized
    debug = False
    print( "build configuration selected: optimized" )
else: # build_type == 'debug':
    my_extra_compile_args_math    = extra_compile_args_math_debug
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_debug
    my_extra_link_args            = extra_link_args_debug
    debug = True
    print( "build configuration selected: debug" )


#########################################################
# Helpers
#########################################################

my_include_dirs = ["."]  # IMPORTANT, see https://github.com/cython/cython/wiki/PackageHierarchy

def ext(extName):
    extPath = extName.replace(".", os.path.sep)+".pyx"
    return Extension( extName,
                      [extPath],
                      extra_compile_args=my_extra_compile_args_nonmath
                    )
def ext_math(extName):
    extPath = extName.replace(".", os.path.sep)+".pyx"
    return Extension( extName,
                      [extPath],
                      extra_compile_args=my_extra_compile_args_math,
                      extra_link_args=my_extra_link_args,
                      libraries=["m"]  # "m" links libm, the math library on unix-likes; see http://docs.cython.org/src/tutorial/external.html
                    )


#########################################################
# Modules
#########################################################

ext_module_ckernel  = ext_math( "cython_kernel" )

#########################################################

setup(
    ext_modules = cythonize( [ ext_module_ckernel ],
                             include_path = my_include_dirs,
                             gdb_debug = debug )
)




# -*- coding: utf-8 -*-
#
# Trying out the NumPy API for Legendre polynomials and Gauss--Legendre quadrature.
#
# JJ 2016-02-16

from __future__ import division, print_function, absolute_import

import numpy as np
import matplotlib.pyplot as plt


def get_legendre_polynomials(max_degree):
    polynomials = []

    # For each degree d, get the polynomial coefficients of a Legendre series
    # that has only the dth degree term. Construct the corresponding Polynomial object,
    #
    # The coefficients are listed from the lowest order to highest.
    #
    for d in range(max_degree):
        # d zeroes followed by a one
        #
        series_coeffs = [ 0. ] * d
        series_coeffs.append( 1. )

        # coefficients for a standard power series 1, x, x**2, ...
        #
        c = np.polynomial.legendre.leg2poly( series_coeffs )

        # create the Polynomial object, remapping the input range to [0,1] for convenience
        #
        polynomials.append( np.polynomial.Polynomial( c, domain=[0., 1.], window=[-1., 1.] ) )

    return polynomials


def main():
    # Set the maximum degree.
    #
    # For our purposes, for accurate results (all polynomials staying within [-1,1] for the whole interval)
    # it seems d = 30 is about the upper limit of what this implementation can do.
    #
    d = 30

    # From the API docs for numpy.polynomial.legendre.leggauss:
    #    Computes the sample points and weights for Gauss-Legendre quadrature.
    #    These sample points and weights will correctly integrate polynomials of degree 2*deg - 1 or less over the interval [-1, 1] with the weight function f(x) = 1.
    #
    # Hence, in Galerkin methods, to exactly handle a mass matrix where neither of the terms is differentiated, using affine mapping to the reference element [0,1]
    # (implying piecewise constant Jacobian), we need to have
    #
    #   2*deg - 1 = 2*d
    #
    # i.e.
    #
    #   deg = (2*d + 1) / 2
    #
#    deg = int(np.ceil( (2*d + 1)/2. ))
#    q,w = np.polynomial.legendre.leggauss( deg )
#    print( deg,(2*deg-1),q,w )

    P = get_legendre_polynomials(d)
    xx = np.linspace(0., 1., 501)

    plt.figure(1)
    plt.clf()
    for p in P:
        plt.plot( xx, p(xx) )

    plt.axis('tight')
    a = plt.axis()
    plt.axis( [ a[0], a[1], a[2]*1.05, a[3]*1.05 ] )

    plt.grid(b=True, which='both')
    plt.title('Legendre polynomials')


    # Try some operations

    # As long as we keep the Polynomial objects, we can multiply them the intuitive way, producing a new Polynomial:
    #
    print( P[2]*P[3] )  # => poly([ 0.    0.75  0.   -3.5   0.    3.75])

    # We can also differentiate them, which is useful for constructing the mass matrix:
    #
    print( P[2].deriv(1)*P[3] )  # => poly([  0.   0.  -9.   0.  15.])

    # Also integration is supported.
    #
    # p.integ() returns the definite integral, as a Polynomial object, from lbnd to an unspecified upper limit x, adding the integration constant k.
    # The value of x is chosen when calling the resulting object.
    #
    # Legendre polynomials are L2-orthogonal on [-1,1]; the scaled ones are orthogonal on [0,1]:
    print( ( (P[2]*P[2]).integ(lbnd=0, k=0) )(1.0) )  # 1/(2 n + 1);  here n = 2, so this = 1/5 = 0.2
    print( ( (P[2]*P[3]).integ(lbnd=0, k=0) )(1.0) )  # zero

    # The integral of  dPn/dx * Pm  over the interval is zero if:
    #
    #  - n + m is even
    #  - n < m  (and by the previous condition, also  n <= m)
    #
    # These observations are based on the L2-orthogonality and the relation
    #
    #   (2 n + 1) P_n = (d/dx)( P_{n+1} - P_{n-1} )         (*)
    #
    # which can be used to get rid of the derivative. The relation (*) follows from Bonnet’s recursion formula,
    #
    #   (n + 1) P_{n+1} = (2 n + 1) P_n - n P_{n-1}
    #
    # By recursive application, (*) leads to the representation
    #
    #   (d/dx) P_{n+1} = (2 n + 1) P_n + ( 2 (n - 2) + 1 ) P_{n-2} + ( 2 (n - 4) + 1 ) P_{n-4} + ...
    #
    # which is guaranteed to bottom out at P_1 and P_0 (by using  P_0 = 1  and  P_1 = x  in (*)).
    #
    # See
    #  https://en.wikipedia.org/wiki/Legendre_polynomials#Additional_properties_of_Legendre_polynomials
    #
    print( ( (P[3].deriv(1)*P[3]).integ(lbnd=0, k=0) )(1.0) )  # zero, n + m even
    print( ( (P[3].deriv(1)*P[1]).integ(lbnd=0, k=0) )(1.0) )  # zero, n + m even
    print( ( (P[2].deriv(1)*P[3]).integ(lbnd=0, k=0) )(1.0) )  # zero, n < m
    print( ( (P[3].deriv(1)*P[2]).integ(lbnd=0, k=0) )(1.0) )  # nonzero (derivative of p3 contains p2, p0)


if __name__ == '__main__':
    main()
    plt.show()




# -*- coding: utf-8 -*-
#
# Trying out the NumPy API for Legendre polynomials and Gauss--Legendre quadrature,
# with an eye toward the modern hierarchical (Lobatto) basis functions for Galerkin methods
# (B. Szabó, I. Babuška, Finite element analysis, John Wiley & Sons, 1991).
#
# JJ 2016-02-16

from __future__ import division, print_function, absolute_import

import time

import numpy as np
import scipy.integrate
import matplotlib.pyplot as plt

import pylu.dgesv as dgesv


class RandomPileOfTestStuff:
    def __init__(self, q=15, tol=1e-8):
        assert( q >= 2 )  # we don't have special case handling for q=1 in build_hierarchical_basis()

        self.q   = q     # max polynomial degree for Legendre polynomials; number of basis functions for hierarchical basis (as in "dG(q)")
        self.tol = tol   # tolerance for nonzero check

        self.P   = None  # Legendre polynomials
        self.N   = None  # hierarchical basis functions (FEM, dG)

        self.C   = None  # dG mass matrix for the first-order problem u' = f(u, t)

        self.get_legendre_polynomials()
        self.build_hierarchical_basis()
        self.dgmass()

    def get_legendre_polynomials(self):
        q = self.q

        P = []

        # For each degree d, get the polynomial coefficients of a Legendre series
        # that has only the dth degree term. Construct the corresponding Polynomial object.
        #
        # The coefficients are listed from the lowest order to highest.
        #
        for d in range(q):
            # d zeroes followed by a one
            #
            series_coeffs = [ 0. ] * d
            series_coeffs.append( 1. )

            # coefficients for a standard power series 1, x, x**2, ...
            #
            c = np.polynomial.legendre.leg2poly( series_coeffs )

            P.append( np.polynomial.Polynomial( c ) )

        self.P = P

    def build_hierarchical_basis(self):
        assert( self.P is not None )

        q = self.q
        P = self.P

        N = []
        N.append( np.polynomial.Polynomial( [0.5, -0.5] ) )  # N_1, will become N[0] in the code, using Polynomial instead of explicit lambda gets us support for .deriv()
        N.append( np.polynomial.Polynomial( [0.5,  0.5] ) )  # N_2
        for j in range(2,q):
            #N.append( np.sqrt( (2.*j - 1.)/2.) * P[j-1].integ(lbnd=-1, k=0) )  # surely this approach makes no numerical sense

            # Explicit solution, using NumPy to evaluate the sum of Legendre polynomials.
            #
            # Much better (and still fast), but not nearly as accurate as evaluating using higher precision internally. See legtest3.py.
            #
            series_coeffs = [ 0. ] * (j-2)
            series_coeffs.extend( [-1., 0., 1.] )  # -P_{j-2} + P_{j}
            c = np.polynomial.legendre.leg2poly( series_coeffs )
            Nj = np.polynomial.Polynomial(c)  /  np.sqrt( 2. * (2.*j - 1.) )
            N.append( Nj )

        self.N = N

    # This numerical approach for generating the matrix is prone to roundoff and obsolete (not to mention stupid
    # since we know that most of the matrix entries should be zero); see the analytical solution in legtest3.py.
    #
    def dgmass(self):
        assert( self.N is not None )

        q = self.q
        N = self.N

        C = np.empty( (q,q), dtype=np.float64 )
        for i in range(q):
            for j in range(q):
                C[i,j] = scipy.integrate.quad( N[j].deriv(1)*N[i], -1., 1. )[0]
        C[ np.abs(C) < self.tol ] = 0.0
        C[0,0] += 1.0  # simulate the effect of the jump term (N_1 is the only function that is nonzero at xi=-1)

        self.C = C


def main():
    # Up to q=24, the full script works despite warnings from quad() in dgmass().
    #
    # For evaluating the hierarchical basis functions only (no dgmass()):
    #
    #   q = 30, still sort of works, small deviations (1e-7) can be seen in the endpoint values of the few highest-order Nj
    #   q = 40, almost works, high-order Nj start getting wobbly
    #   q = 50, completely broken, out of precision
    #
    # By comparison, legtest3.py, which uses SymPy's mpmath (arbitrary precision floating point), works at least up to q=300, but is very slow.
    #
    stuff = RandomPileOfTestStuff(q=24, tol=1e-3)

    # From the API docs for numpy.polynomial.legendre.leggauss:
    #    Computes the sample points and weights for Gauss-Legendre quadrature.
    #    These sample points and weights will correctly integrate polynomials of degree 2*deg - 1 or less over the interval [-1, 1] with the weight function f(x) = 1.
    #
    # Hence, in Galerkin methods, to exactly handle a mass matrix where neither of the terms is differentiated, using affine mapping to the reference element [-1,1]
    # (implying piecewise constant Jacobian), we need to have
    #
    #   2*deg - 1 = 2*d
    #
    # i.e.
    #
    #   deg = (2*d + 1) / 2
    #
#    deg = int(np.ceil( (2*d + 1)/2. ))
#    q,w = np.polynomial.legendre.leggauss( deg )
#    print( deg,(2*deg-1),q,w )

    print( stuff.C )
    print( np.linalg.matrix_rank(stuff.C) )  # should be full rank
    plt.figure(2)
    plt.spy(stuff.C)
    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
#    plt.imshow(M, interpolation="nearest", cmap="Oranges")
#    plt.colorbar()
    plt.title(r"$\mathbf{M}$")


##    L,U,p = dgesv.lup(stuff.C)
##    print( np.transpose(np.nonzero(L)) )
##    print( np.transpose(np.nonzero(U)) )
##    print( p )
##    plt.figure(3)
##    plt.subplot(1,2, 1)
##    plt.spy(L)
##    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
###    plt.imshow(L, interpolation="nearest", cmap="Oranges")
###    plt.colorbar(orientation="horizontal")
##    plt.title(r"$\mathbf{L}$")
##    plt.subplot(1,2, 2)
##    plt.spy(U)
##    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
###    plt.imshow(U, interpolation="nearest", cmap="Oranges")
###    plt.colorbar(orientation="horizontal")
##    plt.title(r"$\mathbf{U}$")


    LU,p = dgesv.lup_packed(stuff.C)
    plt.figure(4)
    plt.spy(LU)
    plt.plot( [0,stuff.q-1], [0,stuff.q-1], 'r--' )
    plt.title(r"$\mathbf{LU}$ (packed format)")

    mincols,maxcols = dgesv.find_bands(LU, 1e-15)
    print( mincols, maxcols )


##    # old Python-based mincols, maxcols finding code
##
##    # Find the smallest column index with nonzero data on each row in L.
##    #
##    # We can use this to "sparsify" the backsolve even though the data structure is dense.
##    #
##    # This assumes that each row has at least one nonzero entry (which is always the case for an invertible matrix).
##    #
##    Lnz = np.nonzero(L)
##    mincols = []
##    rowprev = -1
##    n = len(Lnz[0])
##    i = 0
##    while i < n:
##        if Lnz[0][i] != rowprev:
##            mincols.append(Lnz[1][i])
##            rowprev = Lnz[0][i]
##        i += 1
##    mincols = np.array( mincols, dtype=np.intc, order="C" )
##    print( L )
##    print( mincols )

##    # Find the largest column index with nonzero data on each row in U.
##    #
##    # We can use this to "sparsify" the backsolve even though the data structure is dense.
##    #
##    # This assumes that each row has at least one nonzero entry (which is always the case for an invertible matrix).
##    #
##    Unz = np.nonzero(U)
##    maxcols = []
##    rowprev = -1
##    n = len(Unz[0])
##    i = n - 1
##    while i >= 0:
##        if Unz[0][i] != rowprev:
##            maxcols.append(Unz[1][i])
##            rowprev = Unz[0][i]
##        i -= 1
##    maxcols.reverse()
##    maxcols = np.array( maxcols, dtype=np.intc, order="C" )
##    print( U )
##    print( maxcols )


    # Visualize
    #
    xx = np.linspace(-1., 1., 100001)  # the good thing about the fast approach... smooth curves!
    plt.figure(1)
    plt.clf()
    for func in stuff.N:
        plt.plot( xx, func(xx) )

    plt.axis('tight')
    a = plt.axis()
    plt.axis( [ a[0], a[1], a[2]*1.05, a[3]*1.05 ] )

    plt.grid(b=True, which='both')
    plt.title('Hierarchical basis functions')


    # Try some operations on the original Legendre polynomials
    #
    # As long as we keep the Polynomial objects, we can multiply them the intuitive way, producing a new Polynomial:
    #
    print( stuff.P[2]*stuff.P[3] )  # => poly([ 0.    0.75  0.   -3.5   0.    3.75])

    # We can also differentiate them, which is useful for constructing the mass matrix:
    #
    print( stuff.P[2].deriv(1)*stuff.P[3] )  # => poly([  0.   0.  -9.   0.  15.])

    # Also integration is supported.
    #
    # p.integ() returns the definite integral, as a Polynomial object, from lbnd to an unspecified upper limit x, adding the integration constant k.
    # The value of x is chosen when calling the resulting object.
    #
    # Legendre polynomials are L2-orthogonal on [-1,1]:
    print( ( (stuff.P[2]*stuff.P[2]).integ(lbnd=-1, k=0) )(1.0) )  # 2/(2 n + 1);  here n = 2, so this = 2/5 = 0.4
    print( ( (stuff.P[2]*stuff.P[3]).integ(lbnd=-1, k=0) )(1.0) )  # zero

    # The integral of  dPn/dx * Pm  over the interval is zero if:
    #
    #  - n + m is even
    #  - n < m  (and by the previous condition, also  n <= m)
    #
    # These observations are based on the L2-orthogonality and the relation
    #
    #   (2 n + 1) P_n = (d/dx)( P_{n+1} - P_{n-1} )         (*)
    #
    # which can be used to get rid of the derivative. The relation (*) follows from Bonnet’s recursion formula,
    #
    #   (n + 1) P_{n+1} = (2 n + 1) P_n - n P_{n-1}
    #
    # By recursive application, (*) leads to the representation
    #
    #   (d/dx) P_{n+1} = (2 n + 1) P_n + ( 2 (n - 2) + 1 ) P_{n-2} + ( 2 (n - 4) + 1 ) P_{n-4} + ...
    #
    # which is guaranteed to bottom out at P_1 and P_0 (by using  P_0 = 1  and  P_1 = x  in (*)).
    #
    # See
    #  https://en.wikipedia.org/wiki/Legendre_polynomials#Additional_properties_of_Legendre_polynomials
    #
    print( ( (stuff.P[3].deriv(1)*stuff.P[3]).integ(lbnd=-1, k=0) )(1.0) )  # zero, n + m even
    print( ( (stuff.P[3].deriv(1)*stuff.P[1]).integ(lbnd=-1, k=0) )(1.0) )  # zero, n + m even
    print( ( (stuff.P[2].deriv(1)*stuff.P[3]).integ(lbnd=-1, k=0) )(1.0) )  # zero, n < m
    print( ( (stuff.P[3].deriv(1)*stuff.P[2]).integ(lbnd=-1, k=0) )(1.0) )  # nonzero (derivative of p3 contains p2, p0)


# naive solve (repeat the LU decomposition process each time)
#
def method1(reps, A, b, x):
    for j in range(reps):
#        dgesv.solve( A, b[j,:], x )
        dgesv.solve( A, b, x )

# decompose once, then solve
#
def method2(reps, A, b, x):
    LU,p = dgesv.lup_packed(A)
    for j in range(reps):
#        dgesv.solve_decomposed( LU, p, b[j,:], x )
        dgesv.solve_decomposed( LU, p, b, x )

# decompose once, then solve, utilize banded structure
#
def method3(reps, A, b, x):
    LU,p = dgesv.lup_packed(A)
    mincols,maxcols = dgesv.find_bands(LU, 1e-15)
    for j in range(reps):
#        dgesv.solve_decomposed_banded( LU, p, mincols, maxcols, b[j,:], x )
        dgesv.solve_decomposed_banded( LU, p, mincols, maxcols, b, x )


class MyTimer:
    t0 = None
    l  = None

    def __init__(self, label=""):
        self.label = label

    def __enter__(self):
        self.t0 = time.time()

    def __exit__(self, type, value, traceback):
        dt = time.time() - self.t0
        l  = ("%s: " % self.label) if len(self.label) else "time taken: "
        print( "%s%gs" % (l, dt) )


if __name__ == '__main__':
    main()
    plt.show()

#    # Running the benchmark loop at the Python end makes the banded version look slower (for our matrix M, the C code is actually ~3x faster than the generic non-banded version),
#    # because a large majority of the execution time is taken up by data conversion from Python to C and back (and Python asserts, if enabled).
#    #
#    # To get reliable results on the C code only (which is a realistic use case if used from inside a Cython-accelerated solver, which is the whole point of dgesv.pyx),
#    # the looping must be done inside dgesv.pyx.
#    #
#    reps = 100000
#    for q in range(3, 16):
#        stuff = RandomPileOfTestStuff(q)
#        n = np.shape(stuff.C)[0]
##        b = np.random.uniform(0.0, 1.0, size=(reps,n))  # this makes slicing part of the performance measurement - not good
#        b = np.random.uniform(0.0, 1.0, size=(n,))
#        x = np.empty( [n], dtype=np.float64, order="C" )

#        print( "Timings for %d runs" % reps )
#        with MyTimer("%dx%d naive" % (n,n)) as mt:
#            method1(reps, stuff.C, b, x)
#        with MyTimer("%dx%d decompose-once" % (n,n)) as mt:
#            method2(reps, stuff.C, b, x)
#        with MyTimer("%dx%d decompose-once-banded" % (n,n)) as mt:
#            method3(reps, stuff.C, b, x)




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Classical explicit integrators to update w by one timestep.

These routines use compensated summation to obtain maximal accuracy,
at the cost of 4x math in the final summation step.
"""

from __future__ import division, print_function, absolute_import

import numpy as np

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.types import DTYPE, RTYPE
from pydgq.solver.kernel_interface cimport KernelBase
from pydgq.solver.integrator_interface cimport ExplicitIntegrator
from pydgq.solver.compsum cimport accumulate


cdef class RK4(ExplicitIntegrator):
    def __init__(self, KernelBase rhs):
        """def __init__(self, KernelBase rhs):

Fourth-order Runge-Kutta (RK4).

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ExplicitIntegrator).
"""
        # super
        ExplicitIntegrator.__init__(self, name="RK4", rhs=rhs)

        cdef int n_space_dofs = rhs.n
        self.wrk_arr = np.empty( (5*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef unsigned int j
        cdef int n_space_dofs = self.rhs.n
        cdef DTYPE_t* wstar   = self.wrk  # updated w, used for computing the next k

        # standard RK4 temp variables
        cdef DTYPE_t* k1 = &(self.wrk[n_space_dofs])
        cdef DTYPE_t* k2 = &(self.wrk[2*n_space_dofs])
        cdef DTYPE_t* k3 = &(self.wrk[3*n_space_dofs])
        cdef DTYPE_t* k4 = &(self.wrk[4*n_space_dofs])

        cdef RTYPE_t dtp2  = dt/2.0
        cdef RTYPE_t dtp6  = dt/6.0

        cdef RTYPE_t thalf = t + dtp2
        cdef RTYPE_t tend  = t + dt

        cdef DTYPE_t s
        cdef DTYPE_t c

        # Compute k1, ..., k4
        #
        self.rhs.begin_iteration(0)  # explicit method, single iteration only
        self.rhs.call(w, k1, t)

        for j in range(n_space_dofs):
            wstar[j] = w[j] + dtp2 * k1[j]
        self.rhs.call(wstar, k2, thalf)

        for j in range(n_space_dofs):
            wstar[j] = w[j] + dtp2 * k2[j]
        self.rhs.call(wstar, k3, thalf)

        for j in range(n_space_dofs):
            wstar[j] = w[j] + dt   * k3[j]
        self.rhs.call(wstar, k4, tend)

        # Update
        #
        #   w(new) = w(old) + (dt/6) * ( k1 + 2*k2 + 2*k3 + k4 )
        #
        # Use compensated summation for maximal accuracy. This is important for problems sensitive to initial conditions
        # (such as mechanical problems with low damping).
        #
        for j in range(n_space_dofs):
            s = k1[j]  # the first operand is always exactly representable; here both LHS and RHS are DTYPE_t
            c = 0.0
            accumulate( &s, &c, 2.0*k2[j] )
            accumulate( &s, &c, 2.0*k3[j] )
            accumulate( &s, &c,     k4[j] )

            # we want  w = w + (dt/6) * s,  do the multiplication now
            s *= dtp6
            c *= dtp6

            # s = s + w
            accumulate( &s, &c,      w[j] )
            w[j] = s

        return 1  # explicit method, always one iteration


cdef class RK3(ExplicitIntegrator):
    def __init__(self, KernelBase rhs):
        """def __init__(self, KernelBase rhs):

Kutta's third-order method (commonly known as RK3).

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ExplicitIntegrator).
"""
        # super
        ExplicitIntegrator.__init__(self, name="RK3", rhs=rhs)

        cdef int n_space_dofs = rhs.n
        self.wrk_arr = np.empty( (4*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef unsigned int j
        cdef int n_space_dofs = self.rhs.n
        cdef DTYPE_t* wstar   = self.wrk  # updated w, used for computing the next k

        # RK temp variables
        cdef DTYPE_t* k1 = &(self.wrk[n_space_dofs])
        cdef DTYPE_t* k2 = &(self.wrk[2*n_space_dofs])
        cdef DTYPE_t* k3 = &(self.wrk[3*n_space_dofs])

        cdef RTYPE_t dtp2  = dt/2.0
        cdef RTYPE_t twodt = 2.0*dt
        cdef RTYPE_t dtp6  = dt/6.0

        cdef RTYPE_t thalf = t + dtp2
        cdef RTYPE_t tend  = t + dt

        cdef DTYPE_t s
        cdef DTYPE_t c

        # Compute k1, ..., k3
        #
        self.rhs.begin_iteration(0)  # explicit method, single iteration only
        self.rhs.call(w, k1, t)

        for j in range(n_space_dofs):
            wstar[j] = w[j] + dtp2 * k1[j]
        self.rhs.call(wstar, k2, thalf)

        for j in range(n_space_dofs):
            wstar[j] = w[j]  - dt * k1[j]  + twodt * k2[j]
        self.rhs.call(wstar, k3, tend)

        # Update
        #
        #   w(new) = w(old) + (dt/6) * ( k1 + 4*k2 + k3 )
        #
        # Use compensated summation for maximal accuracy. This is important for problems sensitive to initial conditions
        # (such as mechanical problems with low damping).
        #
        for j in range(n_space_dofs):
            s = k1[j]  # the first operand is always exactly representable; here both LHS and RHS are DTYPE_t
            c = 0.0
            accumulate( &s, &c, 4.0*k2[j] )
            accumulate( &s, &c,     k3[j] )

            # we want  w = w + (dt/6) * s,  do the multiplication now
            s *= dtp6
            c *= dtp6

            # s = s + w
            accumulate( &s, &c,      w[j] )
            w[j] = s

        return 1  # explicit method, always one iteration


cdef class RK2(ExplicitIntegrator):
    def __init__(self, KernelBase rhs, RTYPE_t beta):
        """def __init__(self, KernelBase rhs, RTYPE_t beta):

Parametric second-order Runge-Kutta (RK2).

Parameters:
    beta : RTYPE_t
        Where inside the timestep to take the second evaluation of the RHS.

        beta must be in the half-open interval (0, 1].

        Very small values will cause problems (beta appears in the denominator
        in the final summation formula).

        Popular choices:
            1/2 :
                explicit midpoint method
            2/3 :
                Ralston's method
            1 :
                Heun's method, also known as the explicit trapezoid rule

    other parameters:
        as in ancestor (pydgq.solver.integrator_interface.ExplicitIntegrator).
"""
        # super
        ExplicitIntegrator.__init__(self, name=("RK2 (beta = %g)" % (beta)), rhs=rhs)

        self.beta = beta

        cdef int n_space_dofs = rhs.n
        self.wrk_arr = np.empty( (3*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef unsigned int j
        cdef int n_space_dofs = self.rhs.n
        cdef DTYPE_t* wstar   = self.wrk  # updated w, used for computing the next k

        # RK temp variables
        cdef DTYPE_t* k1 = &(self.wrk[n_space_dofs])
        cdef DTYPE_t* k2 = &(self.wrk[2*n_space_dofs])

        cdef RTYPE_t weight2 = 1.0 / (2.0 * self.beta)
        cdef RTYPE_t weight1 = 1.0 - weight2
        weight1 *= dt
        weight2 *= dt

        cdef RTYPE_t betadt = self.beta*dt
        cdef RTYPE_t tmid   = t + betadt

        cdef DTYPE_t s
        cdef DTYPE_t c

        # Compute k1, k2
        #
        self.rhs.begin_iteration(0)  # explicit method, single iteration only
        self.rhs.call(w, k1, t)

        for j in range(n_space_dofs):
            wstar[j] = w[j] + betadt * k1[j]
        self.rhs.call(wstar, k2, tmid)

        # Update
        #
        #   w(new) = w(old) +  weight1 * k1  +  weight2 * k2
        #
        # Use compensated summation for maximal accuracy. This is important for problems sensitive to initial conditions
        # (such as mechanical problems with low damping).
        #
        for j in range(n_space_dofs):
            s = weight1 * k1[j]  # the first operand is always exactly representable; here both LHS and RHS are DTYPE_t
            c = 0.0
            accumulate( &s, &c, weight2*k2[j] )

            # s = s + w
            accumulate( &s, &c,      w[j] )
            w[j] = s

        return 1  # explicit method, always one iteration


cdef class FE(ExplicitIntegrator):
    def __init__(self, KernelBase rhs):
        """def __init__(self, KernelBase rhs):

Forward Euler.

The most unstable integrator known. Requires a very small timestep. Accuracy is O(dt).

Provided for reference only.

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ExplicitIntegrator).
"""
        # super
        ExplicitIntegrator.__init__(self, name="FE", rhs=rhs)

        cdef int n_space_dofs = rhs.n
        self.wrk_arr = np.empty( (1*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef int n_space_dofs = self.rhs.n
        cdef DTYPE_t* wp      = self.wrk  # w prime (output from RHS)

        self.rhs.begin_iteration(0)  # explicit method, single iteration only
        self.rhs.call(w, wp, t)

        for j in range(n_space_dofs):
            w[j] += dt*wp[j]

        return 1  # explicit method, always one iteration


cdef class SE(ExplicitIntegrator):
    def __init__(self, KernelBase rhs):
        """def __init__(self, KernelBase rhs):

Symplectic Euler.

Accuracy is O(dt).

This method is applicable only to second-order ODE systems
which have been reduced to first order using the companion method.

Let v = u'.

Storage format of w:
    u1, v1, u2, v2, [...], um, vm

where
    m = n / 2

is the size of the original 2nd-order system that was reduced
to the 1st-order system of size n.

Parameters:
    as in ancestor (pydgq.solver.integrator_interface.ExplicitIntegrator).

For a quick explanation of the method, see e.g.:
    http://en.wikipedia.org/wiki/Semi-implicit_Euler_method
    http://www.phy.uct.ac.za/courses/opencontent/phylab2/worksheet9_09.pdf
    http://umu.diva-portal.org/smash/get/diva2:140361/FULLTEXT01.pdf   p.29 ff.

Original reference:
    Niiranen, Jouko: Fast and accurate symmetric Euler algorithm for electromechanical simulations.
    Proceedings of the Electrimacs'99, Sept. 14-16, 1999 Lisboa, Portugal, Vol. 1, pages 71 - 78.
"""
        # super
        ExplicitIntegrator.__init__(self, name="SE", rhs=rhs)

        cdef int n_space_dofs = rhs.n  # this is 2*m
        self.wrk_arr = np.empty( (1*n_space_dofs,), dtype=DTYPE, order="C" )
        self.wrk     = &(self.wrk_arr[0])

    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil:
        cdef unsigned int j
        cdef int n_space_dofs = self.rhs.n
        cdef DTYPE_t* wp      = self.wrk  # w prime (output from RHS)

        self.rhs.begin_iteration(0)  # explicit method, single iteration only
        self.rhs.call(w, wp, t)  # of this, SE uses only the v' components corresponding to the v degrees of freedom

        for j in range(1,n_space_dofs,2):
            w[j] += dt*wp[j]  # qdot, forward diff using current qdotdot
        for j in range(0,n_space_dofs,2):
            w[j] += dt*w[j+1]  # q, backward diff using updated qdot

#        # e.g. for n_space_dofs == 4:
#        w[1] += dt*wp[1]  # v1, forward diff using current v'1
#        w[3] += dt*wp[3]  # v2, forward diff using current v'2
#        w[0] += dt*w[1]   # u1, backward diff using updated u'1 = v1
#        w[2] += dt*w[3]   # u2, backward diff using updated u'2 = v2

        return 1  # explicit method, always one iteration




# -*- coding: utf-8 -*-
"""List-handling utilities."""

from __future__ import division, print_function, absolute_import


def load_balance_list(L, n):
    """Given a list of arbitrary items, split it to n roughly equal-sized parts.

    This is useful for dividing a list of work items in MPI parallelization.
    It is assumed that each work item takes the same amount of time; hence the
    initial distribution is generated by naive integer division.

    If len(L) does not divide evenly with n, the remaining items are distributed
    on an item-by-item basis to the first (len(L) mod n) parts.

    If n > len(L), the items will be distributed on an item-by-item basis to the
    first len(L) parts, and the rest of the parts will get an empty list.
    
    Examples:
        load_balance_list(range(14), 2)
            =>  [[0, 1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12, 13]]
        In this case, the division is even (no remainder).

        load_balance_list(range(15), 2)
            =>  [[0, 1, 2, 3, 4, 5, 6, 14], [7, 8, 9, 10, 11, 12, 13]]
        In this case, the item "14" is left over after the integer division.
        The leftover item is placed in the first part.

        load_balance_list(range(4), 8)
            =>  [[0], [1], [2], [3], [], [], [], []]
        In this case, n is so large that there are not enough items to place
        even one item in each part. The empty list is generated for those parts
        for which no item is available.

    Parameters:
        L = any Python list.

    Return value:
        List of lists: L_out = [L1, L2, L3, ..., Ln]
        where L1, L2, ... are sublists of L.

        It always holds that len(L_out) == n. Note that the lengths of the
        Lj (individual lists inside L_out) may differ by one item, depending on
        whether the division was even.

        If n == 1, the return value is [L] for compatibility
        of the output format with the n > 1 case.

    """
    if n < 1:
        raise ValueError("n must be >= 1; got %d" % (n))

    # If we're splitting L to one part, this is a no-op.
    # But wrap it - the caller is expecting a list of lists.
    #
    if n == 1:
        return [L]

    nitems = len(L)  # number of items to be distributed

    out = []
    if n <= nitems:
        # Items per process.
        blocklen  = nitems // n  # integer division!

        # Leftover items.
        remainder = nitems - blocklen*n  # this is always < n

        # Distribute the items that divided evenly.
        for m in range(n):
            offs = m*blocklen
            out.append( L[offs:offs+blocklen] )

        # Distribute the leftovers (if any).
        if remainder > 0:
            offs = nitems - remainder
            for m in range(remainder):
                out[m].append( L[offs] )
                offs += 1
    else:
        # Distribute one item to each part as long as items are available.
        for m in range(nitems):
            out.append( [L[m]] )  # wrap the item to make a single-item list.

        # Give an empty list to the rest of the parts.
        nempties = n - nitems  # this many empty lists are needed
        for m in range(nempties):
            out.append( [] )

    assert( len(out) == n )
    return out




# -*- coding: utf-8 -*-
"""MPI support wrapper using mpi4py as backend.

License: 2-clause BSD; copyright 2012-2017 Juha Jeronen and University of Jyväskylä.

To load MPI, import this module. The module is designed so that it can
always be imported, regardless of whether mpi4py is installed.

The availability of the mpi4py library can then be queried at runtime
using has_mpi().

Some rudimentary wrappers (get_size(), get_rank(), get_comm_world())
are provided; all the rest should be done manually using mpi4py,
and only if has_mpi() returns True. This way, the same code can
run both with and without MPI.

See also gather_varlength_array(), a buffer-based fast gather
for rank-1 NumPy arrays of varying lengths (like gatherv,
but determines the sizes automatically).

Example:

import mpi_shim

MPI    = mpi_shim.get_mpi()         # either a reference to mpi4py.MPI, or None if n/a
nprocs = mpi_shim.get_size()        # actual value from mpi4py, or 1    if n/a
my_id  = mpi_shim.get_rank()        # actual value from mpi4py, or 0    if n/a
comm   = mpi_shim.get_comm_world()  # actual value from mpi4py, or None if n/a

# The rest works as usual ( adapted from http://mpi4py.scipy.org/docs/usrman/tutorial.html ):

data = (my_id+1)**2  # <insert meaningful calculation here>
if MPI is not None:
   data = comm.gather(data, root=0)

if my_id == 0:
   for i in range(nprocs):
       assert data[i] == (i+1)**2
else:
   assert data is None
"""

from __future__ import division, print_function, absolute_import

import numpy as np  # used in gather_varlength_array()

try:
    import mpi4py.MPI as MPI

    __comm_world = MPI.COMM_WORLD
    __nproc = MPI.Comm.Get_size(__comm_world)
    __my_id = MPI.Comm.Get_rank(__comm_world)

    __library_ok = True

    # Considering the code that uses this module,
    # this is maybe more sensible than always True.
    #
    __mpi_available = (__nproc > 1)
except ImportError:
    __comm_world = None
    __nproc = 1
    __my_id = 0

    __library_ok = False
    __mpi_available = False

# version of mpi_shim
#
# Especially important, since this module is to be copied to each project that uses it.
#
__version__ = "1.0.0"


##########################################
# Wrapper functions
##########################################

def has_mpi4py_library():
    """Return whether the mpi4py library is available (bool).

    This is just a simple library loadedness check.

    If you want to also see whether there are at least 2 MPI processes
    in the current group (and hence parallel processing makes sense),
    use has_mpi() instead.

    """
    return __library_ok

def has_mpi():
    """Return True if the mpi4py library is available and there are at least
    2 MPI processes in the current group. Otherwise return False.

    This is a "does parallel processing make sense?" check.

    See has_mpi4py_library() for just checking library availability.

    """
    return __mpi_available

def get_mpi():
    """Return a reference to the loaded mpi4py.MPI instance.

    If mpi4py is not available, return None.
    """
    if __library_ok:
        return MPI
    else:
        return None

def get_size():
    """Return the size (int) of the MPI group.

    See also:
        get_rank()    (get the ID of the currently running instance)
    """
    return __nproc

def get_rank():
    """Return the MPI rank (int) of the running instance.

    The rank is the identifier of the process, numbered 0, 1, ..., get_size()-1.

    See also:
        get_size()    (get number of running instances)
    """
    return __my_id

def get_comm_world():
    """Return mpi4py.MPI.COMM_WORLD (reference) if mpi4py is available, or None if n/a."""
    return __comm_world


###################################
# MPI-aware printing
###################################

def mpi_print(*args):
    """Print args, but only if running in the root process.

    See also:
        mpi_allprint()    (gather args from all MPI ranks, print in root process)
    """
    if __my_id == 0:  # also true if no MPI
        print(*args)

def mpi_allprint(*args):
    """Gather args from all MPI ranks, print them all in the root process.

    The root process makes a separate call to built-in print for args from each rank.

    MPI rank of the messages is not reported; add this data to the message yourself if desired.

    See also:
        mpi_print()    (print only in the root process)
    """
    comm = get_comm_world()

    if comm is None:  # no MPI
        print(*args)
        return

    allargs = comm.gather(args, root=0)
    if get_rank() == 0:
        for args in allargs:
            print(*args)
    # else no-op


###################################
# Helpers
###################################

def gather_varlength_array(data, datatype, do_allgather=False):
    """Fast (buffer-based) MPI gather for variable-sized data using NumPy arrays.

    Data type of all items must be the same. It must be a primitive datatype
    that mpi4py.MPI.Gather() supports.

    Parameters:
        data : Python list or rank-1 array
            Local data items in each process, to be gathered at the root process.
        datatype : NumPy dtype
            Datatype specification (e.g. np.int, np.float64, np.complex128, ...).
        do_allgather : bool, optional
            If True, do an Allgather() instead of a Gather().

    In different processes, `data` is allowed to have a different number of elements,
    but `datatype` must be the same.

    `datatype` is mandatory so that zero-length input can be handled properly.

    Return value:
        In the root process:
            tuple (data, I), where
                data : rank-1 np.array
                    flattened data array, with data from all processes
                    concatenated in order of increasing MPI rank.
                I : rank-1 np.array of length get_size()+1
                    data from process j is located in data[ I[j] : I[j+1] ].
                    The end fencepost for the last process is provided to avoid
                    the need for special-casing in calling code.

        In all other processes:
            If do_allgather is True, same as in root process, otherwise None.
    """
    # Use an efficient buffer transfer for gathering the data.
    #
    # Because the amount of data per process varies, we must first allgather the data lengths
    # to determine a suitable buffer size. We will use the maximum size in all processes.
    #
    nprocs = get_size()
    data_lengths = np.empty( [nprocs], dtype=int )
    n_local_items = np.size(data)
    n_local_items_array = np.array( n_local_items, dtype=int )  # rank-1 array of length 1

    def lengths_to_offsets(lengths):
        return np.concatenate( ( [0], np.cumsum(lengths) ) )

    comm = get_comm_world()
    comm.Allgather( n_local_items_array, data_lengths )  # order of parms: sendbuf, recvbuf

    # Any leftover array elements are simply unused; we use data_lengths to determine
    # which elements to read.
    #
    max_entries_per_process = np.max(data_lengths)
    nentries = nprocs*max_entries_per_process

    # Check special case: nothing to do if no data
    #
    if max_entries_per_process == 0:
        if get_rank() == 0  or  do_allgather:
            return (np.empty( [0], dtype=datatype ), lengths_to_offsets(data_lengths))
        else:
            return None

    # Prepare receive buffer at root process
    # (or at all processes if allgathering)
    #
    if get_rank() == 0  or  do_allgather:
        data_recvbuf = np.empty( [nentries], dtype=datatype )
    else:
        data_recvbuf = None

    # Prepare and fill send buffers in all processes
    #
    data_sendbuf = np.empty( [max_entries_per_process], dtype=datatype )
    data_sendbuf[0:n_local_items] = np.array( data, dtype=datatype )

    if do_allgather:
        comm.Allgather( data_sendbuf, data_recvbuf )
    else:
        comm.Gather( data_sendbuf, data_recvbuf, root=0 )

    # In the root process, extract the data.
    # (or in all processes if we are allgathering)
    #
    if get_rank() == 0  or  do_allgather:
        result = np.empty( [np.sum(data_lengths)], dtype=datatype )

        # This is similar to util.flatten_list_of_arrays(), but from a single np.array buffer.
        offs_out = 0
        for nproc in range(nprocs):
            offs_in = nproc*max_entries_per_process
            datalen = data_lengths[nproc]
            if datalen > 0:
                result[offs_out:(offs_out+datalen)] = data_recvbuf[offs_in:(offs_in+datalen)]
                offs_out += datalen
        assert( offs_out == np.sum(data_lengths) )
        return (result, lengths_to_offsets(data_lengths))
    else:
        return None

def allgather_varlength_array(data, datatype):
    """Same as gather_varlength_array(data, datatype, do_allgather=True)."""
    return gather_varlength_array(data, datatype, do_allgather=True)




# -*- coding: utf-8 -*-
#
"""Setuptools-based setup script for pydgq."""

from __future__ import division, print_function, absolute_import

import os

# remove file, ignore error if it did not exist
#
# portable version for Python 2.7 and 3.x
#
def myremove(filename):
    try:
        exc = FileNotFoundError  # this does not exist in Python 2.7
    except:
        exc = OSError

    try:
        os.remove( redirect_name )
    except exc:
        pass


#########################################################
# Config
#########################################################

# choose build type here
#
build_type="optimized"
#build_type="debug"


#########################################################
# Init
#########################################################

# check for Python 2.7 or later
# http://stackoverflow.com/questions/19534896/enforcing-python-version-in-setup-py
import sys
if sys.version_info < (2,7):
    sys.exit('Sorry, Python < 2.7 is not supported')

from setuptools import setup
from setuptools.extension import Extension

try:
    from Cython.Build import cythonize
except ImportError:
    sys.exit("Cython not found. Cython is needed to build the extension modules for pydgq.")


#########################################################
# HACK: Attempt to package the correct data file on POSIX
#########################################################

redirect_name = os.path.join("pydgq", "pydgq_data.bin")
file_27_name  = os.path.join("pydgq", "pydgq_data_27.bin")
file_34_name  = os.path.join("pydgq", "pydgq_data_34.bin")

# remove existing symlink or file if any
#
myremove( redirect_name )

# symlink correct file depending on Python version
#
# (FIXME/TODO: for now, we assume that setup.py is running under the same Python version the packaged library will run under)
#
if sys.version_info < (3,0):
    print("Packaging %s as %s" % (file_27_name, redirect_name))
    os.symlink( file_27_name, redirect_name )
else:
    print("Packaging %s as %s" % (file_34_name, redirect_name))
    os.symlink( file_34_name, redirect_name )


#########################################################
# Definitions
#########################################################

extra_compile_args_math_optimized    = ['-march=native', '-O2', '-msse', '-msse2', '-mfma', '-mfpmath=sse']
extra_compile_args_math_debug        = ['-march=native', '-O0', '-g']

extra_compile_args_nonmath_optimized = ['-O2']
extra_compile_args_nonmath_debug     = ['-O0', '-g']

extra_link_args_optimized    = []
extra_link_args_debug        = []


if build_type == 'optimized':
    my_extra_compile_args_math    = extra_compile_args_math_optimized
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_optimized
    my_extra_link_args            = extra_link_args_optimized
    debug = False
    print( "build configuration selected: optimized" )
else: # build_type == 'debug':
    my_extra_compile_args_math    = extra_compile_args_math_debug
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_debug
    my_extra_link_args            = extra_link_args_debug
    debug = True
    print( "build configuration selected: debug" )


#########################################################
# Long description
#########################################################

DESC="""Integrate first-order ODE systems  u'(t) = f(u, t).

The main feature of this library is dG(q), i.e. the
time-discontinuous Galerkin method using a Lobatto basis
(a.k.a. hierarchical polynomial basis).

dG(q) is a very accurate implicit method that often allows
using a rather large timestep. Due to its Galerkin nature,
it also allows inspecting the behavior of the solution
inside the timestep.

Arbitrary q is supported, but often best results are
obtained for q=1 or q=2.

Some classical integrators (RK2, RK3, RK4, IMR, BE)
are also provided for convenience.

The focus is on arbitrary nonlinear problems; all implicit
methods are implemented using fixed-point (Banach/Picard)
iteration.
"""


#########################################################
# Helpers
#########################################################

my_include_dirs = ["."]  # IMPORTANT, see https://github.com/cython/cython/wiki/PackageHierarchy

def ext(extName):
    extPath = extName.replace(".", os.path.sep)+".pyx"
    return Extension( extName,
                      [extPath],
                      extra_compile_args=my_extra_compile_args_nonmath
                    )
def ext_math(extName):
    extPath = extName.replace(".", os.path.sep)+".pyx"
    return Extension( extName,
                      [extPath],
                      extra_compile_args=my_extra_compile_args_math,
                      extra_link_args=my_extra_link_args,
                      libraries=["m"]  # "m" links libm, the math library on unix-likes; see http://docs.cython.org/src/tutorial/external.html
                    )

# http://stackoverflow.com/questions/13628979/setuptools-how-to-make-package-contain-extra-data-folder-and-all-folders-inside
datadirs  = ("doc", "test")
dataexts  = (".py", ".pyx", ".pxd", ".c", ".sh", ".lyx", ".pdf")
datafiles = []
getext = lambda filename: os.path.splitext(filename)[1]
for datadir in datadirs:
    datafiles.extend( [(root, [os.path.join(root, f) for f in files if getext(f) in dataexts])
                       for root, dirs, files in os.walk(datadir)] )

datafiles.append( ('.', ["README.md", "LICENSE.md", "TODO.md", "CHANGELOG.md"]) )

#########################################################
# Modules
#########################################################

ext_module_types      = ext(      "pydgq.solver.types"                )

ext_module_compsum    = ext_math( "pydgq.solver.compsum"              )

ext_module_discontify = ext_math( "pydgq.utils.discontify"            )

ext_module_kernintf   = ext_math( "pydgq.solver.kernel_interface"     )
ext_module_bkernels   = ext_math( "pydgq.solver.builtin_kernels"      )

ext_module_intgintf   = ext_math( "pydgq.solver.integrator_interface" )
ext_module_explicit   = ext_math( "pydgq.solver.explicit"             )
ext_module_implicit   = ext_math( "pydgq.solver.implicit"             )
ext_module_galerkin   = ext_math( "pydgq.solver.galerkin"             )

ext_module_odesolve   = ext_math( "pydgq.solver.odesolve"             )

#########################################################

# Extract __version__ from the package __init__.py
# (since it's not a good idea to actually run __init__.py during the build process).
#
# http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
#
import ast
with open('pydgq/__init__.py') as f:
    for line in f:
        if line.startswith('__version__'):
            version = ast.parse(line).body[0].value.s
            break
    else:
        version = '0.0.unknown'
        print( "WARNING: Version information not found, using placeholder '%s'" % (version) )


setup(
    name = "pydgq",
    version = version,
    author = "Juha Jeronen",
    author_email = "juha.jeronen@jyu.fi",
    url = "https://github.com/Technologicat/pydgq",

    description = "ODE system solver using dG(q) (time-discontinuous Galerkin w/ Lobatto basis)",
    long_description = DESC,

    license = "BSD",
    platforms = ["Linux"],  # free-form text field; http://stackoverflow.com/questions/34994130/what-platforms-argument-to-setup-in-setup-py-does

    classifiers = [ "Development Status :: 4 - Beta",
                    "Environment :: Console",
                    "Intended Audience :: Developers",
                    "Intended Audience :: Science/Research",
                    "License :: OSI Approved :: BSD License",
                    "Operating System :: POSIX :: Linux",
                    "Programming Language :: Cython",
                    "Programming Language :: Python",
                    "Programming Language :: Python :: 2",
                    "Programming Language :: Python :: 2.7",
                    "Programming Language :: Python :: 3",
                    "Programming Language :: Python :: 3.4",
                    "Topic :: Scientific/Engineering",
                    "Topic :: Scientific/Engineering :: Mathematics",
                    "Topic :: Software Development :: Libraries",
                    "Topic :: Software Development :: Libraries :: Python Modules"
                  ],

    setup_requires = ["cython", "numpy"],
    # pydgq.utils.precalc can optionally use mpi4py, but it is not mandatory.
    # Also, thre no need to run precalc in common use cases, since we include
    # the data file pydgq_data.bin (computed using default options) in the package.
    # Thus, we just leave out mpi4py.
    install_requires = ["numpy", "pylu"],
    provides = ["pydgq"],

    # same keywords as used as topics on GitHub
    keywords = ["numerical integration ordinary-differential-equations ode ivp ode-solver solver galerkin discontinuous-galerkin cython numpy"],

    ext_modules = cythonize( [ ext_module_types,
                               ext_module_compsum,
                               ext_module_discontify,
                               ext_module_kernintf, ext_module_bkernels,
                               ext_module_intgintf, ext_module_explicit, ext_module_implicit, ext_module_galerkin,
                               ext_module_odesolve,  ],
                             include_path = my_include_dirs,
                             gdb_debug = debug ),

    # Declare packages so that  python -m setup build  will copy .py files (especially __init__.py).
    packages = ["pydgq", "pydgq.solver", "pydgq.utils"],

    # Install also Cython headers so that other Cython modules can cimport ours
    # FIXME: force sdist, but sdist only, to keep the .pyx files (this puts them also in the bdist)
    package_data={'pydgq':        ['*.pxd', '*.pyx', '*.bin'],  # note: paths relative to each package
                  'pydgq.solver': ['*.pxd', '*.pyx'],
                  'pydgq.utils':  ['*.pxd', '*.pyx']},

    # Disable zip_safe, because:
    #   - Cython won't find .pxd files inside installed .egg, hard to compile libs depending on this one
    #   - dynamic loader may need to have the library unzipped to a temporary folder anyway (at import time)
    zip_safe = False,

    # Usage examples; not in a package
    data_files = datafiles
)


# remove symlink created earlier
#
myremove( redirect_name )




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Helper for processing C^{-1} (finitely discontinuous) functions for correct plotting (i.e. without drawing a vertical line at the discontinuities).

Effectively, this is equivalent to the following NumPy snippet, using the idea from

    http://stackoverflow.com/questions/10377593/how-to-drop-connecting-lines-where-the-function-is-discontinuous

def discontify( data, idxs, fill="nan" ):  # data, idxs: rank-1 np.arrays
    # prepare
    nz = idxs.tolist()
    nz.sort()
    nz.reverse()

    if len(nz) == 0:
        return data

    # process
    out = data.copy()
    if fill == "nan":
        for pos in nz:
            out = np.insert(out, pos+1, np.nan)
    elif fill == "prev":
        for pos in nz:
            out = np.insert(out, pos+1, out[pos])
    else:
        raise ValueError("Unknown fill '%s'; available: 'nan', 'prev'" % fill)

    return out

but (even if the preparation was done only once) this Cython version works much faster when idxs contains a lot of elements.
"""

from __future__ import division, print_function, absolute_import

import numpy as np

from pydgq.solver.types cimport DTYPE_t
from pydgq.solver.types  import DTYPE, DNAN


def discontify( DTYPE_t[::1] data, int[::1] idxs, str fill="nan" ):
    """Make a rank-1 np.array discontinuous by inserting fill values.

    This is useful for plotting C^{-1} functions correctly (i.e. without drawing a vertical line at the discontinuities).

    The implementation uses Cython and runs at native (C) speed, so if there are a lot of discontinuities (as in e.g. a dG numerical solution),
    it is much faster than using np.insert() for each discontinuity individually.

    Parameters:
        data : rank-1 np.array
            input data
        idxs : rank-1 np.array
            indices in original data *after which* a copy of the fill value is to be inserted. Must be already sorted in increasing order.
        fill : str, one of:
            "nan"  :
                fill by NaN. Useful for y data of a C^{-1} function, as it disables plotting (in Matplotlib) for that data point.
            "prev" :
                fill by copying the element at the immediately preceding position in data. Useful for x data of a C^{-1} function,
                as Matplotlib will check the x values for their increasingness and emits a runtime warning if there are any NaNs.

    Returns:
        rank-1 np.array
            the discontified data. Note that the indexing no longer matches the original data (due to the inserted fill values).

    """
    cdef unsigned int ndata, nidxs
    ndata = data.shape[0]
    nidxs = idxs.shape[0]

    if nidxs == 0:  # if no idxs, this is a no-op
        return np.asanyarray(data)

    cdef unsigned int j, k
    cdef unsigned int offs = 0
    cdef DTYPE_t nan = DNAN
    cdef DTYPE_t[::1] out = np.empty( (ndata + nidxs,), dtype=DTYPE, order="C" )

    DEF MODE_NAN  = 1
    DEF MODE_PREV = 2
    cdef int fill_mode
    if fill == "nan":
        fill_mode = MODE_NAN
    elif fill == "prev":
        fill_mode = MODE_PREV
    else:
        raise ValueError("Unknown fill '%s'; available: 'nan', 'prev'" % (fill))

    with nogil:
        # copy data up to first discontinuity
        for k in range(idxs[0]+1):
            out[k] = data[k]

        # add the fill value
        if fill_mode == MODE_NAN:
            out[idxs[0]+1] = nan
        else: # fill_mode == MODE_PREV:
            out[idxs[0]+1] = out[idxs[0]]

        offs += 1  # one discontinuity handled; increase orig -> output index offset by one

        # handle the rest of discontinuities (if > 1 specified)
        if nidxs > 1:
            for j in range(1, nidxs):
                # copy data between discontinuities j-1 and j
                for k in range(idxs[j-1]+1, idxs[j]+1):
                    out[k+offs] = data[k]

                # add the fill value
                if fill_mode == MODE_NAN:
                    out[idxs[j]+offs+1] = nan
                else: # fill_mode == MODE_PREV:
                    out[idxs[j]+offs+1] = out[idxs[j]+offs]

                offs += 1  # one more discontinuity handled

        # copy data after the last discontinuity
        for k in range(idxs[nidxs-1]+1, ndata):
            out[k+offs] = data[k]

    return np.asanyarray(out)




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
"""Compensated summation (Kahan algorithm), Python/Cython.

This Python interface exposes compensated cumulative summation of rank-1 arrays.
"""

from __future__ import division, print_function, absolute_import

from pydgq.solver.types cimport RTYPE_t, ZTYPE_t
from pydgq.solver.types import RTYPE, ZTYPE

import numpy as np


#########################################################################################
# Python interface
#########################################################################################

# These are provided to produce accurate cumulative sums.
# Not currently used by pydgq.

def cumsum1d_compensated( data ):
    """def cumsum1d_compensated( data ):

Like np.cumsum(), but using compensated summation (Kahan algorithm).

Implemented only for rank-1 np.arrays of dtypes double (np.float64) and double complex (np.complex128).
"""
    cdef RTYPE_t[::1] inr, outr
    cdef ZTYPE_t[::1] inz, outz

    if isinstance(data, np.ndarray):
        if data.ndim != 1:
            raise ValueError("This function supports only rank-1 arrays, got an array of rank %d" % (data.ndim))

        if data.dtype == ZTYPE:
            inz  = data
            outz = np.empty_like(data)
            cs1dz( &inz[0], &outz[0], data.shape[0] )  # modifies outz in-place
            return np.asanyarray(outz)
        elif data.dtype == RTYPE:
            inr  = data
            outr = np.empty_like(data)
            cs1dr( &inr[0], &outr[0], data.shape[0] )  # modifies outr in-place
            return np.asanyarray(outr)
        else:
            raise TypeError("Unsupported dtype '%s' for cumsum1d_compensated() with np.ndarray; valid: %s, %s" % (data.dtype, ZTYPE, RTYPE))
    else:
        raise TypeError("Unsupported argument type '%s' for cumsum1d_compensated(); %s" % (type(data), np.ndarray))


#########################################################################################
# Cython interface
#########################################################################################

# data : in, values to sum (length n)
# out  : out, cumulative sum (length n-1)
# n    : length of input data
#
cdef void cs1dr( RTYPE_t* data, RTYPE_t* out, unsigned int n ) nogil:
    cdef RTYPE_t s = data[0]
    cdef RTYPE_t c = 0.0
    cdef unsigned int j
    out[0] = s
    for j in range(1,n):
        accumulate( &s, &c, data[j] )
        out[j] = s

# data : in, values to sum (length n)
# out  : out, cumulative sum (length n-1)
# n    : length of input data
#
cdef void cs1dz( ZTYPE_t* data, ZTYPE_t* out, unsigned int n ) nogil:
    cdef ZTYPE_t s = data[0]
    cdef ZTYPE_t c = 0.0
    cdef unsigned int j
    out[0] = s
    for j in range(1,n):
        accumulatez( &s, &c, data[j] )
        out[j] = s




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t
from pydgq.solver.kernel_interface cimport KernelBase
from pydgq.solver.integrator_interface cimport ImplicitIntegrator

# Implicit midpoint rule
cdef class IMR(ImplicitIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil

# Backward Euler (implicit Euler)
cdef class BE(ImplicitIntegrator):
    cdef int call(self, DTYPE_t* w, RTYPE_t t, RTYPE_t dt) nogil




# -*- coding: utf-8 -*-
#
# Interface for Cython and Python based computational kernels for evaluating the RHS f() in  w' = f(w).

# cpdef methods cannot be nogil, so we must work around as follows:
#
# KernelBase --> cdef void call(...) nogil, abstract method for interface (NOTE: this class **does not** define callback())
#   CythonKernel --> implement call(), provide hook for Cython cdef callback(...) nogil
#   PythonKernel --> implement call(), provide hook for Python def callback(...), which our call() invokes in a "with gil" block

from __future__ import absolute_import

from pydgq.solver.types cimport DTYPE_t, RTYPE_t

cdef class KernelBase:
    cdef DTYPE_t* w    # old state vector (memory owned by caller)
    cdef DTYPE_t* out  # new state vector (memory owned by caller)
    cdef int n         # n_space_dofs
    cdef int timestep
    cdef int iteration

    # interface for solver
    #
    cdef void begin_timestep(self, int timestep) nogil
    cdef void begin_iteration(self, int iteration) nogil
    cdef void call(self, DTYPE_t* w, DTYPE_t* out, RTYPE_t t) nogil  # interface for solver (abstract method)

# Base class for kernels implemented in Cython.
#
# Cython kernels will run in nogil mode.
#
cdef class CythonKernel(KernelBase):
    cdef void call(self, DTYPE_t* w, DTYPE_t* out, RTYPE_t t) nogil  # interface for solver (implemented here)
    cdef void callback(self, RTYPE_t t) nogil  # hook for user code

# Base class for kernels implemented in pure Python.
#
# Python kernels will acquire the gil for calling callback().
#
cdef class PythonKernel(KernelBase):
    cdef DTYPE_t[::1] w_arr    # Python-accessible view into w   (provided by __getattr__ as "w")
    cdef DTYPE_t[::1] out_arr  # Python-accessible view into out (provided by __getattr__ as "out")

    cdef void call(self, DTYPE_t* w, DTYPE_t* out, RTYPE_t t) nogil  # interface for solver (implemented here)
    # here callback() is a Python (regular def) function  # hook for user code




# -*- coding: utf-8 -*-
#
# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True
#
from __future__ import division, print_function, absolute_import

import numpy as np

from pydgq.solver.types cimport RTYPE_t, DTYPE_t
from pydgq.solver.types  import RTYPE,   DTYPE
from pydgq.solver.kernel_interface cimport CythonKernel

from libc.math cimport sin, cos, log, exp, sqrt

cdef extern from "math.h":
    double M_PI

# A simple cosine kernel with phase-shifted components.
#
# See python_kernel_test.py for a pure-Python version of this.
#
# Note that Cython requires us to put the cdef class declarations into cython_kernel.pxd.
#
# The custom kernel only needs to override callback(); even __init__ is not strictly needed,
# unless adding some custom parameters (like here).
#
cdef class MyKernel(CythonKernel):
    def __init__(self, n, omega):  # omega : rad/s
        # super
        CythonKernel.__init__(self, n)

        # custom init
        self.omega = omega

    # implementation of callback()
    #
    cdef void callback(self, RTYPE_t t) nogil:
        cdef RTYPE_t phi0_j
        cdef int j
        for j in range(self.n):
            phi0_j = (<RTYPE_t>(j+1) / self.n) * 2. * M_PI
            self.out[j] = cos(phi0_j + self.omega*t)

    # known analytical solution, for testing the integrators
    #
    # (only needed by the testing script; usual custom kernels do not have this part)

    # one component
    def __sol(self, RTYPE_t[::1] tt, RTYPE_t phi0):
        cdef int nt = tt.shape[0]
        cdef DTYPE_t[::1] out = np.empty( (nt,), dtype=DTYPE, order="C" )
        cdef int j
        for j in range(nt):
            out[j] = 1./self.omega * sin(phi0 + self.omega*tt[j])
        return np.asanyarray(out)  # np.ndarray needed to support broadcast

    # all components
    def reference_solution(self, RTYPE_t[::1] tt):
        cdef int nt = tt.shape[0]
        cdef DTYPE_t[:,::1] ww = np.empty( (nt,self.n), dtype=DTYPE, order="C" )
        cdef RTYPE_t[::1] zero = np.array( (0.,), dtype=RTYPE, order="C" )  # the tt input to __sol() must be an array
 
        cdef DTYPE_t[::1] tmp
        cdef RTYPE_t phi0_j
        cdef int j, k
        for j in range(self.n):
            phi0_j = (<RTYPE_t>(j+1) / self.n) * 2. * M_PI
            tmp = self.__sol(tt,phi0_j) - self.__sol(zero,phi0_j)  # shift to account for the initial condition (all solution components start at zero)
            for k in range(nt):
                ww[k,j] = tmp[k]

        return np.asanyarray(ww)

